from bifabrik.cfg.engine.Configuration import Configuration
from bifabrik.cfg.engine.Configuration import CfgProperty

class FileSourceConfiguration(Configuration):
    """Configuration related to loading data from files (aspects common for all file formats)
    """
    def __init__(self):
        self._explicitProps = {}
        self.__encoding = None
        self.__locale = None
        self.__moveFilesToArchive: bool = False
        self.__archiveFolder: str = None
        self.__archiveFilePattern: str = '{filename}_{timestamp}{extension}'


    @CfgProperty
    def encoding(self) -> str:
        """From pandas:
        How encoding errors are treated. https://docs.python.org/3/library/codecs.html#standard-encodings
        """
        return self.__encoding
    @encoding.setter(key='encoding')
    def encoding(self, val):
        self.__encoding = val

    @CfgProperty
    def locale(self) -> str:
        """From spark:
        sets a locale as language tag in IETF BCP 47 format. If None is set, it uses the default value, en-US. For instance, locale is used while parsing dates and timestamps.
        """
        return self.__locale
    @locale.setter(key='locale')
    def jsonLocale(self, val):
        self.__locale = val

    @CfgProperty
    def moveFilesToArchive(self) -> bool:
        """True / False whether to move loaded source files to archive after the pipeline finishes succesfully.
            The archiveFolder and archiveFilePattern need to be configured for this to work.
            Default False
        """
        return self.__moveFilesToArchive
    @moveFilesToArchive.setter(key='moveFilesToArchive')
    def moveFilesToArchive(self, val):
        self.__moveFilesToArchive = val
    
    @CfgProperty
    def archiveFolder(self) -> str:
        """The folder where to archive processed source files (in the source lakehouse).
        """
        return self.__archiveFolder
    @archiveFolder.setter(key='archiveFolder')
    def archiveFolder(self, val):
        self.__archiveFolder = val
    
    @CfgProperty
    def archiveFilePattern(self) -> str:
        """The file pattern for archiving processed source files. 
        Supported placeholders:
        {filename} : source file name without suffix
        {extension} : source file name extension, including the '.'
        {timestamp} : current timestamp as '%Y_%m_%d_%H_%M_%S_%f'

        Default '{filename}_{timestamp}{extension}'
        """
        return self.__archiveFilePattern
    @archiveFilePattern.setter(key='archiveFilePattern')
    def archiveFilePattern(self, val):
        self.__archiveFilePattern = val



"""
File system utilities - glob pattern search, Fabric path normalization and more
"""

import notebookutils.mssparkutils.fs
import glob2
import regex
import sempy.fabric as spf
import logging
import os
import datetime
import time

__mounts = None
__defaultWorkspaceId = spf.get_notebook_workspace_id()
__defaultWorkspaceRefName = 'default'
__defaultLakehouseId = None
__defaultLakehouseWorkspaceId = None
__defaultLakehouseRefName = 'default'
#lgr = lg.getLogger()


def normalizeFileApiPath(path: str):
    """Normalizes a file path to the form of "/lakehouse/default/Files/folder/..."
    """
    # used internally for logging to a mounted directory (int the notebook's default lakehouse)
    r = path
    if not r.startswith('/'):
        r = '/' + r

    lhpt = '/lakehouse'
    dfpt = '/default'
    fpt = '/files'
    lp = r.lower()
    if lp.startswith(lhpt):
        lp = lp[len(lhpt):]
        r = r[len(lhpt):]
    if lp.startswith(dfpt):
        lp = lp[len(dfpt):]
        r = r[len(dfpt):]
    if lp.startswith(fpt):
        lp = lp[len(fpt):]
        r = r[len(fpt):]
    
    r = '/lakehouse/default/Files' + r
    return r

def normalizeRelativeSparkPath(path: str):
    """Normalizes a file path to the form of "Files/folder/file.csv"
    """
    r = path
    if not r.startswith('/'):
        r = '/' + r
    
    lhpt = '/lakehouse'
    dfpt = '/default'
    fpt = '/files'
    lp = r.lower()
    
    lp = r.lower()
    if lp.startswith(lhpt):
        lp = lp[len(lhpt):]
        r = r[len(lhpt):]
    if lp.startswith(dfpt):
        lp = lp[len(dfpt):]
        r = r[len(dfpt):]
    if lp.startswith(fpt):
        lp = lp[len(fpt):]
        r = r[len(fpt):]
    
    r = 'Files' + r
    return r

# abfss://6dcac488-f099-451a-91f1-0945791bf22c@onelake.dfs.fabric.microsoft.com/a2939674-cc1c-4814-9094-8c16e5682502/Files/CsvFiles
def normalizeAbfsPath(path: str, lakehouseBasePath: str, inFiles = False, inTables = False):
    """Normalizes a path to the form of "abfss://{workspaceid}@onelake.dfs.fabric.microsoft.com/{lakehouseid}/Files/CsvFiles or .../Tables/Table1"
    """
    r = path

    if not r.startswith('/'):
        r = '/' + r
    
    lhpt = '/lakehouse'
    dfpt = '/default'
    fpt = '/files'
    lp = r.lower()
    
    lp = r.lower()
    if lp.startswith('abfss:'):
        return r
    if not r.startswith('/'):
        r = '/' + r
        lp = '/' + lp
    
    if inFiles and not (r.startswith('/Files')):
        r = '/Files' + r
    elif inTables and not (r.startswith('/Tables')):
        r = '/Tables' + r
    
    wPrefix = f'{lakehouseBasePath}{r}'
    return wPrefix


def normalizeAbfsFilePath(path: str, lakehouseBasePath: str):
    """Normalizes a path to the form of "abfss://{workspaceid}@onelake.dfs.fabric.microsoft.com/{lakehouseid}/Files/CsvFiles"
    You can use getLakehousePath() to get the lakehouse path as "abfss://{workspaceid}@onelake.dfs.fabric.microsoft.com"
    """
    return normalizeAbfsPath(path, lakehouseBasePath=lakehouseBasePath, inFiles = True)

def normalizeAbfsTablePath(path: str, lakehouseBasePath: str):
    """Normalizes a path to the form of "abfss://{workspaceid}@onelake.dfs.fabric.microsoft.com/{lakehouseid}/Tables/Table1"
    """
    return normalizeAbfsPath(path, lakehouseBasePath=lakehouseBasePath, inTables = True)

def filePatternSearch(path: str, lakehouse: str = None, workspace: str = None, useImplicitDefaultLakehousePath = False) -> list[str]:
    """Searches the Files/ directory of the current lakehouse
    using glob to match patterns. Returns the list of files as ABFS paths.

    Examples
    --------
    >>> bifabrik.utils.fsUtils.filePatternSearch("fld1/*/data/*.csv")
    ...     ["abfss://...@onelake.dfs.fabric.microsoft.com/.../Files/fld1/subf1/data/file11.csv", "abfss://...@onelake.dfs.fabric.microsoft.com/.../Files/fld1/subf2/data/file22.csv"]

    param useImplicitDefaultLakehousePath indicates whether '/lakehouse/default' should be used to refer to the default lakehouse
    """
    
    
    log = logging.getLogger('bifabrik')
    res = []
    #pathNorm = normalizeRelativeSparkPath(path)

    lhPath = getLakehousePath(lakehouse = lakehouse, workspace = workspace, useImplicitDefaultLakehousePath=useImplicitDefaultLakehousePath)
    pathNorm = normalizeAbfsFilePath(path, lhPath)
    pathNormTrimSuffix = pathNorm[len(lhPath) + len('/Files/'):] #pathNorm[len('Files/'):]
    pathPts = pathNormTrimSuffix.split("/")
    startPath = lhPath + '/Files'
    defaultLh = '/lakehouse/default/'
    isInDefaultLakehouse = False
    if startPath.startswith(defaultLh):
        startPath = startPath[len(defaultLh):]
        isInDefaultLakehouse = True
    searchLocations = [startPath]

    if len(pathPts) == 0:
        return res
    
    for i in range(len(pathPts)):
        pathPt = pathPts[i]
        if len(searchLocations) == 0:
            return res
        nextLevel = []
        for location in searchLocations:
            log.info(f'Searching location {location}')
            if not notebookutils.mssparkutils.fs.exists(location):
                warn = f'Location {location} does not exist'
                print(warn)
                log.warning(warn)
                continue
            subLocations = notebookutils.mssparkutils.fs.ls(location)
            subLocationNames = [fi.name for fi in subLocations]
            subLocationsFilteredT = glob2.fnmatch.filter(subLocationNames, pathPt, True, False, None)
            subLocationsFiltered = [x[0] for x in subLocationsFilteredT] 
            #print(subLocationsFiltered)
            subLocationsDict = {}
            for sbl in subLocations:
                subLocationsDict[sbl.name] = sbl
            #print(subLocationsDict)
            for slf in subLocationsFiltered:
                #print(slf)
                finfo = subLocationsDict[slf]
                slfPath = location + '/' + slf
                if len(pathPts) == i + 1:
                    if finfo.isFile:
                        if isInDefaultLakehouse:
                            res.append(f'{defaultLh}{slfPath}')
                        else:
                            res.append(slfPath)
                elif (finfo.isDir):
                    nextLevel.append(slfPath)
        if len(pathPts) == i + 1:
            return res
        else:
            searchLocations = nextLevel

def fileExists(path: str, lakehouse: str = None, workspace: str = None) -> bool:
    searchResults = filePatternSearch(path = path, lakehouse = lakehouse, workspace = workspace)
    if len(searchResults) > 0:
        return True
    return False

def getMounts():
    global __mounts
    if __mounts is None:
        __mounts = notebookutils.mssparkutils.fs.mounts()
    return __mounts

def getDefaultLakehouseAbfsPath() -> str:
    for mp in getMounts():
        if mp.mountPoint == '/default':
            # print(f"Default Lakehouse is: {mp.source}")
            return mp.source
        return None
    
def isGuid(uuid):
    UUID_PATTERN = regex.compile(r'^[\da-f]{8}-([\da-f]{4}-){3}[\da-f]{12}$', regex.IGNORECASE)
    #print(uuid)
    #print(type(uuid))
    r = bool(UUID_PATTERN.match(uuid))
    return r

def getWorkspaceAndLakehouseIdFromMountSource(mountSource):
    m = regex.match(r"abfss://(.+)@onelake.dfs.fabric.microsoft.com/(.+)", mountSource)
    return { 'workspaceId': m.captures(1)[0], 'lakehouseId': m.captures(2)[0] }

class LakehouseMeta:
    lakehouseName: str = None
    lakehouseId: str = None
    workspaceName: str = None
    workspaceId: str = None
    basePath: str = None

class LakehouseMap():

    def __init__(self, workspaceId: str, workspaceName: str):
        self.__dictById = {}
        self.__dictByName = {}
        self.workspaceId = workspaceId
        self.workspaceName = workspaceName
    
    def addLakehouse(self, lh: LakehouseMeta):
        self.__dictById[lh.lakehouseId] = lh
        self.__dictByName[lh.lakehouseName.lower()] = lh

    def __getitem__(self, key) -> LakehouseMeta:
        if isGuid(key):
            if key in self.__dictById:
                return self.__dictById[key]
            return None
        if key.lower() in self.__dictByName:
            return self.__dictByName[key.lower()]
        return None

    def __contains__(self,key):
        if self.__getitem__(key) is not None:
            return True
        return False
    
    def __iter__(self):
        for k in self.__dictById:
            yield self.__dictById[k]

class WorkspaceMap:

    def __init__(self):
        self.__dictById = {}
        self.__dictByName = {}
    
    def addWorkspace(self, lm: LakehouseMap):
        self.__dictById[lm.workspaceId] = lm
        self.__dictByName[lm.workspaceName.lower()] = lm

    def __getitem__(self, key) -> LakehouseMap:
        #print(key)
        #print(self.__dictByName)
        if isGuid(key):
            if key in self.__dictById:
                return self.__dictById[key]
            return None
        if key.lower() in self.__dictByName:
            return self.__dictByName[key.lower()]
        return None
    
    def __contains__(self,key):
        if self.__getitem__(key) is not None:
            return True
        return False
    
    def __iter__(self):
        for k in self.__dictById:
            yield self.__dictById[k]

__workspaceMap = WorkspaceMap()

defaultMount = getDefaultLakehouseAbfsPath()
if defaultMount is not None:
    wsLh = getWorkspaceAndLakehouseIdFromMountSource(defaultMount)    
    __defaultLakehouseId = wsLh['lakehouseId']
    __defaultLakehouseWorkspaceId = wsLh['workspaceId']

def mapLakehouses(workspaceId: str, workspaceName: str):
    lm = LakehouseMap(workspaceId = workspaceId, workspaceName = workspaceName)
    #__workspaceMap.addWorkspace(lm)
    #print(workspaceId)
    #print(workspaceName)
    #print('----')
    lhs = notebookutils.mssparkutils.lakehouse.list(workspaceId)
    for lh in lhs:
        l = LakehouseMeta()
        l.lakehouseName = lh.displayName
        l.lakehouseId = lh.id
        l.workspaceName = workspaceName
        l.workspaceId = workspaceId
        l.basePath = f"abfss://{workspaceId}@onelake.dfs.fabric.microsoft.com/{lh.id}"
        lm.addLakehouse(l)
    return lm

def mapWorkspaces():
    global __workspaceMap
    global __defaultWorkspaceId
    global __defaultWorkspaceRefName
    global __defaultLakehouseId
    global __defaultLakehouseRefName

    wss = spf.list_workspaces()
    #display(wss)
    #print(type(wss[wss.Type == 'Workspace']))
    for wix, ws in wss[wss.Type == 'Workspace'].iterrows():
        lm = mapLakehouses(workspaceId = ws['Id'], workspaceName = ws['Name'])
        __workspaceMap.addWorkspace(lm)

    if __defaultWorkspaceId not in __workspaceMap:
        mwslm = mapLakehouses(workspaceId = __defaultWorkspaceId, workspaceName = 'My workspace')
        __workspaceMap.addWorkspace(mwslm)
    
    return __workspaceMap

mapWorkspaces()


def getLakehouseMeta(lakehouse: str, workspace: str = None, suppressNotFound = False) -> LakehouseMeta:
    """
    Finds the lakehouse amoung available workspaces and returns its metadata
        abfss://{workspace id}@onelake.dfs.fabric.microsoft.com/{lakehouse id}/Files
    
    :param lakehouse: the lakehouse name or ID (if None, the current notebook's deault lakehouse is used)
    :param workspace: the name or ID of the workspace containg the lakehouse (if None, the workspace of the current notebook is used)
    
    :return: LakehouseMeta
    """
    global __workspaceMap
    global __defaultWorkspaceId
    global __defaultLakehouseId

    global __defaultWorkspaceRefName
    global __defaultLakehouseRefName

    # use the workspace of the notebook
    if workspace is None:
        workspace = __defaultWorkspaceId
    if workspace == __defaultLakehouseRefName:
        workspace = __defaultWorkspaceId

    # use the notebook's default lakehouse
    if lakehouse is None:
        lakehouse = __defaultLakehouseId
    if lakehouse == __defaultLakehouseRefName:
        lakehouse = __defaultLakehouseId

    errMsg = f'Could not find lakehouse {lakehouse} in workspace {workspace}'

    if lakehouse is None:
        if suppressNotFound:
            return None
        raise Exception(errMsg)
    
    lhMap = __workspaceMap[workspace]
    if lhMap is None:
        if suppressNotFound:
            return None
        raise Exception(errMsg)
    
    lh = lhMap[lakehouse]
    if lh is None:
        if suppressNotFound:
            return None
        raise Exception(errMsg)
    return lh

def currentLakehouse() -> LakehouseMeta:
    '''Info on the notebook's default (attached) lakehouse
    '''
    l =  getLakehouseMeta(lakehouse = __defaultLakehouseId, workspace = __defaultLakehouseWorkspaceId, suppressNotFound = True)
    if l is None: 
        return None
    return {
        'lakehouseName' : l.lakehouseName,
        'lakehouseId' : l.lakehouseId,
        'workspaceName' : l.workspaceName,
        'workspaceId' : l.workspaceId,
        'basePath' : l.basePath
    } 

def currentLakehouseName() -> str:
    '''Name of the notebook's default (attached) lakehouse
    '''
    clh = currentLakehouse()
    if clh is None:
        return None
    return clh['lakehouseName']

def getLakehousePath(lakehouse: str, workspace: str = None, suppressNotFound = False, useImplicitDefaultLakehousePath = False):
    """
    Returns the lakehouse path as abfss://{workspace id}@onelake.dfs.fabric.microsoft.com/{lakehouse id}
    it can then be appended as e.g
        abfss://{workspace id}@onelake.dfs.fabric.microsoft.com/{lakehouse id}/Tables
        abfss://{workspace id}@onelake.dfs.fabric.microsoft.com/{lakehouse id}/Files
    
    :param lakehouse: the lakehouse name or ID (if None, the current notebook's deault lakehouse is used)
    :param workspace: the name or ID of the workspace containg the lakehouse (if None, the workspace of the current notebook is used)
    """
    meta = getLakehouseMeta(lakehouse, workspace, suppressNotFound)
    if meta is None:
        if suppressNotFound:
            return None
        errMsg = f'Could not find lakehouse {lakehouse} in workspace {workspace}'
        raise Exception(errMsg)
    
    if meta.workspaceId == __defaultWorkspaceId and meta.lakehouseId == __defaultLakehouseId and useImplicitDefaultLakehousePath:
        return '/lakehouse/default'

    return meta.basePath

def archiveFiles(files, archiveFolder, filePattern, lakehouse = None, workspace = None):
    if len(files) == 0:
        return
    if archiveFolder is None:
        raise Exception(f'Cannot archive the file {files[0]}, as the archive folder has not been set')
    if filePattern is None:
        raise Exception(f'Cannot archive the file {files[0]}, as the archive file patter has not been set')
    
    # ABFS path to the archive folder
    lhPath = getLakehousePath(lakehouse = lakehouse, workspace = workspace)
    folderPathNorm = normalizeAbfsFilePath(archiveFolder, lhPath) #.removesuffix('/')
    if not notebookutils.mssparkutils.fs.exists(folderPathNorm):
        notebookutils.mssparkutils.fs.mkdirs(folderPathNorm)

    for file in files:
        log = logging.getLogger('bifabrik')

        fname = os.path.basename(file)
        rootName, extension = os.path.splitext(fname)

        ct = datetime.datetime.now()
        cts = ct.strftime("%Y_%m_%d_%H_%M_%S_%f")
        
        fullPattern = '{folder}/' + filePattern
        fullTargetPath = fullPattern.format(folder = folderPathNorm, filename = rootName, timestamp = cts, extension = extension)

        log.info(f'Archiving {file} -> {fullTargetPath}')
        notebookutils.mssparkutils.fs.mv(file, fullTargetPath)

def getWorkspaceId(workspace):
    if workspace is None:
        return __defaultWorkspaceId
    lhMap = __workspaceMap[workspace]
    if lhMap is None:
        return None
    return lhMap.workspaceId


from bifabrik.cfg.engine.Configuration import Configuration
from bifabrik.cfg.engine.Configuration import CfgProperty

class ExcelConfiguration(Configuration):
    """Configuration related to loading Excel files
    """
    def __init__(self):
        self._explicitProps = {}
        self.__sheetName = 0
        self.__header = 0
        self.__names = None
        self.__usecols = None
        self.__skiprows = None
        self.__nrows = None
        self.__thousands = None
        self.__decimal = '.'

    @CfgProperty
    def sheetName(self):
        """From pandas:
        Strings are used for sheet names. Integers are used in zero-indexed sheet positions (chart sheets do not count as a sheet position). Lists of strings/integers are used to request multiple sheets. Specify None to get all worksheets.
        Available cases:
        Defaults to 0: 1st sheet as a DataFrame
        1: 2nd sheet as a DataFrame
        "Sheet1": Load sheet with name “Sheet1”
        [0, 1, "Sheet5"]: Load first, second and sheet named “Sheet5” as a dict of DataFrame
        None: All worksheets

        default: 0
        """
        return self.__sheetName
    @sheetName.setter(key='sheetName')
    def sheetName(self, val):
        self.__sheetName = val

    @CfgProperty
    def header(self):
        """From pandas:
        Row (0-indexed) to use for the column labels of the parsed DataFrame. If a list of integers is passed those row positions will be combined into a MultiIndex. Use None if there is no header.
        
        default: 0
        """
        return self.__header
    @header.setter(key='header')
    def header(self, val):
        self.__header = val

    @CfgProperty
    def names(self):
        """From pandas:
        List of column names to use. If file contains no header row, then you should explicitly pass header=None.
        
        default: None
        """
        return self.__names
    @names.setter(key='names')
    def names(self, val):
        self.__names = val

    @CfgProperty
    def usecols(self):
        """From pandas:
        If None, then parse all columns.
        If str, then indicates comma separated list of Excel column letters and column ranges (e.g. “A:E” or “A,C,E:F”). Ranges are inclusive of both sides.
        If list of int, then indicates list of column numbers to be parsed (0-indexed).
        If list of string, then indicates list of column names to be parsed.
        If callable, then evaluate each column name against it and parse the column if the callable returns True.
        Returns a subset of the columns according to behavior above.

        default: None
        """
        return self.__usecols
    @usecols.setter(key='usecols')
    def usecols(self, val):
        self.__usecols = val

    @CfgProperty
    def skiprows(self):
        """From pandas:
        Line numbers to skip (0-indexed) or number of lines to skip (int) at the start of the file. If callable, the callable function will be evaluated against the row indices, returning True if the row should be skipped and False otherwise. An example of a valid callable argument would be lambda x: x in [0, 2].
        
        default: None
        """
        return self.__skiprows
    @skiprows.setter(key='skiprows')
    def skiprows(self, val):
        self.__skiprows = val

    @CfgProperty
    def nrows(self):
        """From pandas:
        Number of rows to parse.
        
        default: None
        """
        return self.__nrows
    @nrows.setter(key='nrows')
    def nrows(self, val):
        self.__nrows = val

    @CfgProperty
    def thousands(self):
        """From pandas:
        Thousands separator for parsing string columns to numeric. Note that this parameter is only necessary for columns stored as TEXT in Excel, any numeric columns will automatically be parsed, regardless of display format.
        
        default: None
        """
        return self.__thousands
    @thousands.setter(key='thousands')
    def thousands(self, val):
        self.__thousands = val

    @CfgProperty
    def decimal(self):
        """From pandas:
        Character to recognize as decimal point for parsing string columns to numeric. Note that this parameter is only necessary for columns stored as TEXT in Excel, any numeric columns will automatically be parsed, regardless of display format.(e.g. use ‘,’ for European data).
        
        default: '.'
        """
        return self.__decimal
    @decimal.setter(key='decimal')
    def decimal(self, val):
        self.__decimal = val


from bifabrik.cfg.engine.ConfigContainer import ConfigContainer
from bifabrik.cfg.specific.SecurityConfiguration import SecurityConfiguration
from bifabrik.cfg.DataSourceConfigurationBase import DataSourceConfigurationBase

class SharePointListSourceConfiguration(DataSourceConfigurationBase):

    def __init__(self):
        self.__security = SecurityConfiguration()
        super().__init__()
        

    @property
    def security(self) -> SecurityConfiguration:
        return self.__security


from bifabrik.src.DataSource import DataSource
from pyspark.sql.dataframe import DataFrame
import bifabrik.utils.log as lg
import pyodbc
import pandas as pd
import notebookutils.mssparkutils.fs
import notebookutils.mssparkutils.credentials

class WarehouseSqlSource(DataSource):
    """TSQL data source loading data from a Fabric data warehouse
    Service principal authentication is used to connect to the warehouse. This needs to be configured as below

    Examples
    --------
    > import bifabrik as bif
    >
    > bif.config.security.keyVaultUrl = 'https://kv-contoso.vault.azure.net/'
    > bif.config.security.servicePrincipalClientId = '56712345-1234-7890-abcd-abcd12344d14'
    > bif.config.security.servicePrincipalClientSecretKVSecretName = 'contoso-clientSecret'
    > bif.config.sourceStorage.sourceWarehouseConnectionString = 'dxtxxxxxxbue.datawarehouse.fabric.microsoft.com'
    >
    > bif.fromWarehouseSql('''
    > SELECT CountryOrRegion, Year, PublicHolidayCount FROM [DW1].[dbo].[HolidayCountsYearly]
    > ''').toTable('HolidayCountsYearlyFromDW').run()
    """

    def __init__(self, parentPipeline, query):
        super().__init__(parentPipeline)
        self.__query = query
        self.__config = None
        
    
    def __str__(self):
        return f'Warehouse SQL source: {self.__query}'
    
    def execute(self, input) -> DataFrame:
        lgr = lg.getLogger()
        self.__logger = lgr
        self._error = None
        
        config = self._pipeline.configuration
        self.__config = config
        
        # connect ODBC
        odbcServer = config.sourceStorage.sourceWarehouseConnectionString
        odbcDatabase = config.sourceStorage.sourceWarehouseName
        odbcTimeout = config.sourceStorage.sourceWarehouseConnectionTimeout
        principalClientId = config.security.servicePrincipalClientId

        if odbcServer is None:
            raise Exception('Cannot connect to warehouse - sourceStorage.sourceWarehouseConnectionString not configured (copy this from the properties of your warehouse)')
        if odbcDatabase is None:
            raise Exception('Cannot connect to warehouse - sourceStorage.sourceWarehouseName not configured')
        if principalClientId is None:
            raise Exception('Cannot connect to warehouse - security.servicePrincipalClientId not configured (service principal authentication is used)')
        if config.security.keyVaultUrl is None:
            raise Exception('Cannot connect to warehouse - security.keyVaultUrl not configured (service principal authentication is used and the client secret needs to be stored in key vault)')
        if config.security.servicePrincipalClientSecretKVSecretName is None:
            raise Exception('Cannot connect to warehouse - security.servicePrincipalClientSecretKVSecretName not configured (service principal authentication is used and the client secret needs to be stored in key vault)')
        
        principalClientSecret = notebookutils.mssparkutils.credentials.getSecret(config.security.keyVaultUrl, config.security.servicePrincipalClientSecretKVSecretName)
        constr = f"driver=ODBC Driver 18 for SQL Server;server={odbcServer};database={odbcDatabase};UID={principalClientId};PWD={principalClientSecret};Authentication=ActiveDirectoryServicePrincipal;Encrypt=yes;Timeout={odbcTimeout};"
        if odbcDatabase is None:
            constr = f"driver=ODBC Driver 18 for SQL Server;server={odbcServer};UID={principalClientId};PWD={principalClientSecret};Authentication=ActiveDirectoryServicePrincipal;Encrypt=yes;Timeout={odbcTimeout};"
        self.__odbcConnection = pyodbc.connect(constr)

        pd_df = pd.read_sql(self.__query, self.__odbcConnection)
        df = self._spark.createDataFrame(pd_df)
        
        self._result = df

        self.__odbcConnection.close()
        self._completed = True


from bifabrik.src.DataSource import DataSource
from pyspark.sql.dataframe import DataFrame

class SparkSqlSource(DataSource):
    """SQL (SparkSQL) data source - executes a SparkSQL query

    Examples
    --------
    > import bifabrik as bif
    >
    > bif.fromSql.query('SELECT OrderId, CustomerId  FROM LH1.FactOrderLine WHERE ProductId = 791057').toTable('TransformedOrders1').run()
    """
    
    def __init__(self, parentPipeline):
        super().__init__(parentPipeline)
        self._query = ""
    
    def __str__(self):
        return f'SQL source: {self._query}'
    
    def query(self, query: str):
        """The source SQL query (SparkSQL) to be executed against the current lakehouse
        """
        self._query = query
        return self
    
    def execute(self, input) -> DataFrame:
        df = self._spark.sql(self._query)
        self._result = df
        self._completed = True


from bifabrik.cfg.engine.Configuration import Configuration
from bifabrik.cfg.engine.Configuration import CfgProperty

class SourceStorageConfiguration(Configuration):
    """Defines the data source lakehouse or warehouse. By default refers to the default lakehouse of the notebook.
    """
    def __init__(self):
        self._explicitProps = {}
        self.__sourceWorkspace = None
        self.__sourceLakehouse = None
        self.__sourceWarehouseConnectionString = None
        self.__sourceWarehouseConnectionTimeout = 600
        self.__sourceWarehouseName = None

    @CfgProperty
    def sourceWorkspace(self) -> str:
        """The workspace from which pipeline sources load data - ID or name
        """
        return self.__sourceWorkspace
    @sourceWorkspace.setter(key='sourceWorkspace')
    def sourceWorkspace(self, val):
        self.__sourceWorkspace = val

    @CfgProperty
    def sourceLakehouse(self) -> str:
        """The lakehouse from which pipeline sources load data - ID or name
        """
        return self.__sourceLakehouse
    @sourceLakehouse.setter(key='sourceLakehouse')
    def sourceLakehouse(self, val):
        self.__sourceLakehouse = val

    @CfgProperty
    def sourceWarehouseConnectionString(self) -> str:
        """Connection string to the source Fabric warehouse (server name like dxt....datawarehouse.fabric.microsoft.com)
        """
        return self.__sourceWarehouseConnectionString
    @sourceWarehouseConnectionString.setter(key='sourceWarehouseConnectionString')
    def sourceWarehouseConnectionString(self, val):
        self.__sourceWarehouseConnectionString = val

    @CfgProperty
    def sourceWarehouseConnectionTimeout(self) -> str:
        """Warehouse ODBC connection timetout in seconds
        default 600
        """
        return self.__sourceWarehouseConnectionTimeout
    @sourceWarehouseConnectionTimeout.setter(key='sourceWarehouseConnectionTimeout')
    def sourceWarehouseConnectionTimeout(self, val):
        self.__sourceWarehouseConnectionTimeout = val

    @CfgProperty
    def sourceWarehouseName(self) -> str:
        """Name of the warehouse from which the pipelines read data
        """
        return self.__sourceWarehouseName
    @sourceWarehouseName.setter(key='sourceWarehouseName')
    def sourceWarehouseName(self, val):
        self.__sourceWarehouseName = val


import datetime
from pyspark.sql.functions import *

def addNARecord(df, spark, targetTableName: str, identityColumn:str ):
        na_data = []
        col_names = []
        casts = []

        for f in df.dtypes:
            col_name = f[0]
            col_type = f[1]
            
            if col_name == identityColumn:
                na_data.append(-1)
            elif col_type == 'string' or col_type.startswith('char') or col_type.startswith('varchar'):
                na_data.append('N/A')
            elif col_type == 'date' or col_type == 'timestamp':
                na_data.append('2000-01-01')
            else:
                na_data.append(0)
            casts.append(col(col_name).cast(col_type))
            col_names.append(col_name)
        
        na_df = spark.createDataFrame([tuple(na_data)],col_names)
        cast_df = na_df.select(*casts)
        union_df = cast_df.union(df)
        return union_df
        


from bifabrik.cfg.engine.Configuration import Configuration
from bifabrik.cfg.engine.Configuration import CfgProperty

class ValidationConfiguration(Configuration):
    """Configuration related to saving data to delta tables in a Fabric lakehouse or warehouse
    """
    def __init__(self):
        self._explicitProps = {}
        self.__resultColumnName = 'ValidationResult'
        self.__messageColumnName = 'ValidationMessage'
        self.__errorResultValue = 'Error'
        self.__warningResultValue = 'Warning'
        # fail / log / None
        self.__onError = 'fail'
        self.__onWarning = 'log'
        self.__testName = None
        self.__maxErrors = 10
        self.__maxWarnings = 10
        
    @CfgProperty
    def resultColumnName(self) -> str:
        """Column of the input dataframe that contains the validation result. The recognized values of the column are set in the configuration - see
        errorResultValue, warningResultValue, okResultValue

        default 'ValidationResult'
        """
        return self.__resultColumnName
    @resultColumnName.setter(key='resultColumnName')
    def resultColumnName(self, val):
        self.__resultColumnName = val

    @CfgProperty
    def messageColumnName(self) -> str:
        """Column of the input dataframe that contains the validation message. This can explain the reason for validation errors. 
        It will be included in the log / exception thrown from errors / warnings.

        default 'ValidationMessage'
        """
        return self.__messageColumnName
    @messageColumnName.setter(key='messageColumnName')
    def messageColumnName(self, val):
        self.__messageColumnName = val

    @CfgProperty
    def errorResultValue(self) -> str:
        """If the configured {resultColumnName} in the input data has the {errorResultValue}, the row will be considered a validation error.

        default 'Error'
        """
        return self.__errorResultValue
    @errorResultValue.setter(key='errorResultValue')
    def errorResultValue(self, val):
        self.__errorResultValue = val

    @CfgProperty
    def warningResultValue(self) -> str:
        """If the configured {resultColumnName} in the input data has the {warningResultValue}, the row will be considered a validation warning.

        default 'Warning'
        """
        return self.__warningResultValue
    @warningResultValue.setter(key='warningResultValue')
    def warningResultValue(self, val):
        self.__warningResultValue = val

    @CfgProperty
    def onError(self) -> str:
        """What to do when a validation error is encountered.
        Fail = throw exception and stop execution
        Log = log the warning to console and log file
        Nothing / None = ignore

        default 'Fail'
        """
        return self.__onError
    @onError.setter(key='onError')
    def onError(self, val):
        self.__onError = val

    @CfgProperty
    def onWarning(self) -> str:
        """What to do when a validation warning is encountered
        Fail = throw exception and stop execution
        Log = log the warning to console and log file
        Nothing / None = ignore

        default 'Log'
        """
        return self.__onWarning
    @onWarning.setter(key='onWarning')
    def onWarning(self, val):
        self.__onWarning = val

    @CfgProperty
    def testName(self) -> str:
        """Name of the current validation test - will be included in logs / exceptions
        """
        return self.__testName
    @testName.setter(key='testName')
    def testName(self, val):
        self.__testName = val

    @CfgProperty
    def maxErrors(self) -> int:
        """Maximum number of errors to be reported
        
        default 10
        """
        return self.__maxErrors
    @maxErrors.setter(key='maxErrors')
    def maxErrors(self, val):
        self.__maxErrors = val

    @CfgProperty
    def maxWarnings(self) -> int:
        """Maximum number of warnings to be reported

        default 10
        """
        return self.__maxWarnings
    @maxWarnings.setter(key='maxWarnings')
    def maxWarnings(self, val):
        self.__maxWarnings = val



from bifabrik.dst.DataDestination import DataDestination
from bifabrik.base.Pipeline import Pipeline
from bifabrik.cfg.TableDestinationConfiguration import TableDestinationConfiguration
from pyspark.sql.dataframe import DataFrame as SparkDf
from bifabrik.utils import fsUtils
from bifabrik.utils import log

class SparkDfDestination(DataDestination):
    """Saves data to a spark dataframe.

    Examples
    --------
    > import bifabrik as bif
    >
    > df = bif.fromSql('SELECT * FROM SomeTable').toSparkDf().run()
    """
    def __init__(self, pipeline: Pipeline):
        super().__init__(pipeline)

    def __str__(self):
        return f'Spark DF destination'
    
    def execute(self, input: SparkDf) -> None:
        self._result = input
        self._error = None
        self._completed = True


from bifabrik.base.Task import Task

class DataDestination(Task):
    def __init__(self, parentPipeline):
        super().__init__(parentPipeline)

    def save(self) -> None:
        """Save the data to the destination. This is the "run" method at the end of the chain. 
        Returns the result of the last task, if any.
        """
        self._pipeline.execute()


"""
Utilities using the Fabric REST API:
 - running pipelines
"""


import bifabrik.utils.fsUtils as fsu
import bifabrik.utils.log as lg
import bifabrik as bif
import sempy.fabric as spf
import time

def executePipeline(pipeline: str, workspace: str = None, parameters: dict = {}, waitForExit: bool = False):
    """Execute a Fabric pipeline

    Parameters:
        pipeline (str):     Name or ID of the pipeline
        workspace (str):    Name or ID of the workspace where the pipeline is located (default None = current workspace)
        parameters (dict):  Dictionary of parameters to pass to the pipeline (default {})
        waitForExit (bool): Wait for the pipeline to exit (default False)
    """
    workspaceId = fsu.getWorkspaceId(workspace)
    if workspaceId is None:
        raise Exception(f'The workspace {workspace} was not found')
    
    pipelineId = None
    client = spf.FabricRestClient()

    if fsu.isGuid(pipeline):
        pipelineId = pipeline
    else:
        its = client.get(f'https://api.fabric.microsoft.com/v1/workspaces/{workspaceId}/items?type=DataPipeline')
        for i in its.json()['value']:
            if i['displayName'] == pipeline:
                pipelineId = i['id']
        
        if pipelineId is None:
            raise Exception(f'Pipeline {pipeline} not found')
        
        jobType = f"Pipeline"
        relativeUrl = f"/v1/workspaces/{workspaceId}/items/{pipelineId}/jobs/instances?jobType={jobType}"
        print(relativeUrl)
        response = None
        if len(parameters) == 0:
            response = client.post(relativeUrl)
        else:
            payload = {
                "executionData": {
                    "parameters": parameters
                }    
            }
            response = client.post(relativeUrl, json = payload)
        
        location = response.headers['Location']

        progress = client.get(location).json()
        print(progress)

        if not waitForExit:
            return progress
        
        status = progress['status']

        waitTime = 5
        totalWait = 0
        while status in ['NotStarted', 'InProgress']:
            time.sleep(waitTime)
            totalWait = totalWait + waitTime
            if totalWait > 5:
                waitTime = waitTime * 2
            progress = client.get(location).json()
            status = progress['status']
            print(f'After {totalWait} s: {status}')
        
        if status == 'Completed':
            return
        
        raise Exception(f'Pipeline ended with status {progress}')
        
    #  {'id': 'a7385224-4b19-48de-99e6-e02c8077efbc',
    #  'itemId': '0e56ed7f-ccfc-4d1f-a5d6-dad14b8ebb79',
    #  'jobType': 'Pipeline',
    #  'invokeType': 'Manual',
    #  'status': 'InProgress',
    #  'failureReason': None,
    #  'rootActivityId': 'fa2c78a5-c643-41bd-b35c-4630a37dbc6a',
    #  'startTimeUtc': '2024-05-26T11:08:13.95',
    #  'endTimeUtc': None}

def executePipelineWait(pipeline: str, workspace: str = None, parameters: dict = {}):
    """Execute a Fabric pipeline and wait for it to exit

    Parameters:
        pipeline (str):     Name or ID of the pipeline
        workspace (str):    Name or ID of the workspace where the pipeline is located (default None = current workspace)
        parameters (dict):  Dictionary of parameters to pass to the pipeline (default {})
    """
    executePipeline(pipeline = pipeline, workspace = workspace, parameters = parameters, waitForExit = True)





from bifabrik.dst.DataDestination import DataDestination
from bifabrik.base.Pipeline import Pipeline
from bifabrik.cfg.TableDestinationConfiguration import TableDestinationConfiguration
from pyspark.sql.dataframe import DataFrame
from bifabrik.utils import fsUtils
import bifabrik.utils.log as lg
from bifabrik.utils import fsUtils
#from pyspark.sql.functions import col
from pyspark.sql.functions import *
from pyspark.sql.window import Window
import time
import datetime
import notebookutils.mssparkutils.fs
from bifabrik.utils import tableUtils as tu
import uuid
import pyodbc
import pandas as pd
import bifabrik.dst.CommonDestinationUtils as commons

# reuse the table destination configuration
class WarehouseTableDestination(DataDestination, TableDestinationConfiguration):
    """Saves data to a warehouse table. Uses pyodbc with service principal authentication, which needs to be configured.
    See https://rjankovic.github.io/bifabrik/tutorial/dst_warehouse_table.html

    Examples
    --------
    > import bifabrik as bif
    >
    > bif.config.security.servicePrincipalClientId = '56712345-1234-7890-abcd-abcd12344d14'
    > bif.config.security.keyVaultUrl = 'https://kv-contoso.vault.azure.net/'
    > bif.config.security.servicePrincipalClientSecretKVSecretName = 'contoso-clientSecret'
    > bif.config.destinationStorage.destinationWarehouseName = 'WH_GOLD'
    > bif.config.destinationStorage.destinationWarehouseConnectionString = 'dxtxxxxxxbue.datawarehouse.fabric.microsoft.com'
    >
    > bif.fromSql('''
    > SELECT countryOrRegion `CountryOrRegion`
    > ,YEAR(date) `Year` 
    > ,COUNT(*) `PublicHolidayCount`
    > FROM LH_SILVER.publicholidays
    > GROUP BY countryOrRegion
    > ,YEAR(date)
    > ''').toWarehouseTable('HolidayCountsYearly').run()
    """
    def __init__(self, pipeline: Pipeline, targetTableName: str, targetSchemaName: str = 'dbo'):
        super().__init__(pipeline)
        TableDestinationConfiguration.__init__(self)
        self.__targetTableName = targetTableName
        self.__targetSchemaName = targetSchemaName
        self.__data = None
        self.__config = None
        self.__lhBasePath = None
        self.__identityColumn = None
        self.__insertDateColumn = None
        self.__tableExists = False
        self.__logger = None
        self.__tempTableName = None
        self.__tempTableLocation = None
        self.__odbcConnection = None
        self.__watermarkColumn = None
        self.__addNARecord = False

    def __str__(self):
        return f'Warehouse table destination: [{self.__targetSchemaName}].[{self.__targetTableName}]'
    
    def execute(self, input: DataFrame) -> None:
        lgr = lg.getLogger()
        self.__logger = lgr
        self._error = None
        
        if input is None:
            msg = f'The table destination `{self.__targetTableName}` has no input; terminating'
            print(msg)
            lgr.warning(msg)
            return
        
        self.__data = input

        mergedConfig = self._pipeline.configuration.mergeToCopy(self)
        self.__config = mergedConfig
        self.__tableConfig = mergedConfig.destinationTable
        self.__databaseName = mergedConfig.destinationStorage.destinationWarehouseName

        incrementMethod = mergedConfig.destinationTable.increment
        if incrementMethod is None:
            incrementMethod = 'overwrite'
        self.__incrementMethod = incrementMethod
        
        # connect ODBC
        odbcServer = mergedConfig.destinationStorage.destinationWarehouseConnectionString
        odbcDatabase = mergedConfig.destinationStorage.destinationWarehouseName
        odbcTimeout = mergedConfig.destinationStorage.destinationWarehouseConnectionTimeout
        principalClientId = mergedConfig.security.servicePrincipalClientId
        
        if odbcServer is None:
            raise Exception('Cannot connect to warehouse - destinationStorage.destinationWarehouseConnectionString not configured (copy this from the properties of your warehouse)')
        if odbcDatabase is None:
            raise Exception('Cannot connect to warehouse - destinationStorage.destinationWarehouseName not configured')
        if principalClientId is None:
            raise Exception('Cannot connect to warehouse - security.servicePrincipalClientId not configured (service principal authentication is used)')
        if mergedConfig.security.keyVaultUrl is None:
            raise Exception('Cannot connect to warehouse - security.keyVaultUrl not configured (service principal authentication is used and the client secret needs to be stored in key vault)')
        if mergedConfig.security.servicePrincipalClientSecretKVSecretName is None:
            raise Exception('Cannot connect to warehouse - security.servicePrincipalClientSecretKVSecretName not configured (service principal authentication is used and the client secret needs to be stored in key vault)')
        
        principalClientSecret = notebookutils.mssparkutils.credentials.getSecret(mergedConfig.security.keyVaultUrl, mergedConfig.security.servicePrincipalClientSecretKVSecretName)
        constr = f"driver=ODBC Driver 18 for SQL Server;server={odbcServer};database={odbcDatabase};UID={principalClientId};PWD={principalClientSecret};Authentication=ActiveDirectoryServicePrincipal;Encrypt=yes;Timeout={odbcTimeout};"
        self.__odbcConnection = pyodbc.connect(constr)

        # create schema if not exists
        findSchemaQuery = f"SELECT COUNT(*) FROM sys.schemas WHERE [name] = '{self.__targetSchemaName}'"
        findSchemaDf = self.__execute_select(findSchemaQuery)
        schemaCount = findSchemaDf[0][0]
        if schemaCount == 0:
            self.__execute_dml(f"CREATE SCHEMA [{self.__targetSchemaName}]")

        # check if table exists
        findTableDf = self.__execute_select(f'''
        SELECT COUNT(*)
        FROM sys.tables t 
        INNER JOIN sys.schemas s ON t.schema_id = s.schema_id
        WHERE s.name = '{self.__targetSchemaName}' AND t.name = '{self.__targetTableName}'
        ''')

        tableCount = findTableDf[0][0]
        if tableCount == 0:
            self.__tableExists = False
        else:
            self.__tableExists = True


        # watermark filter filters self.__data DF - this has to be done before thedata is written to the lakehouse temp table
        self.__filterByWatermark()
        # the same replacement as in a lakehouse needs to be applied to the column in order to save the temp table to the lakehouse
        self.__replaceInvalidCharactersInColumnNames()
        

        # save to temp table in the lakehouse
        dstLh = mergedConfig.destinationStorage.destinationLakehouse
        dstWs = mergedConfig.destinationStorage.destinationWorkspace
        self.__lhBasePath = fsUtils.getLakehousePath(dstLh, dstWs)
        if self.__lhBasePath is None:
            raise Exception(f'''The warehouse destination needs to use a warehouse for temporary data storage. 
                            Either connect the notebook to a lakehouse or configure the destinationLakehouse and destinationWorkspace properties in bifabrik.config.destinationStorage.''')
        self.__lhMeta = fsUtils.getLakehouseMeta(dstLh, dstWs)
        self.__destinationLakehouse = self.__lhMeta.lakehouseName
        self.__tempTableName = f"temp_{self.__targetTableName}_{str(uuid.uuid4())}".replace('-', '_')
        self.__tempTableLocation = self.__lhBasePath + "/Tables/" + self.__tempTableName
        if dstLh is None:
            self.__tempTableLocation = "Tables/" + self.__tempTableName
        self.__data.write.mode("overwrite").format("delta").option("overwriteSchema", "true").save(self.__tempTableLocation)
        # sleep for 3 s until the lakehouse table comes online
        # LESS THAN IDEAL
        time.sleep(3)

        # determine column types
        inputTableColumns = []

        # add identity column to table structure if specified
        identityColumnPattern = self.__tableConfig.identityColumnPattern
        if identityColumnPattern is not None:
            self.__identityColumn = self.__tableConfig.identityColumnPattern.format(tablename = self.__targetTableName, databasename = self.__databaseName)
            inputTableColumns.insert(0, [self.__identityColumn, 'BIGINT'])
        
        # ordinary columns...
        dts = self.__data.dtypes
        for sc in dts:
            col_type = 'VARCHAR(8000)'
            s_name = sc[0]
            s_type = sc[1]
            if s_type == 'string' or s_type.startswith('char') or s_type.startswith('varchar'):
                col_type = 'VARCHAR(8000)'
            elif s_type == 'long' or s_type == 'bigint':
                col_type = 'BIGINT'
            elif s_type.startswith('bool'):
                col_type = 'BIT'
            elif s_type.startswith('int'):
                col_type = 'INT'
            elif s_type == 'tinyint' or s_type == 'byte' or s_type == 'smallint' or s_type == 'short':
                col_type = 'SMALLINT'
            elif s_type == 'double':
                col_type = 'FLOAT'
            elif s_type == 'float' or s_type == 'real':
                col_type = 'FLOAT'
            elif s_type == 'date' or s_type == 'timestamp':
                col_type = 'DATETIME2(6)'
            elif s_type.startswith('binary') or s_type.startswith('byte'):
                col_type = 'VARBINARY(8000)'
            elif s_type.startswith('decimal('):
                col_type = s_type.replace('decimal', 'DECIMAL')
            
            inputTableColumns.append([s_name, col_type])

        # add insert date column
        insertDateColumn = self.__tableConfig.insertDateColumn
        if insertDateColumn is not None:
            self.__insertDateColumn = insertDateColumn
            inputTableColumns.append([self.__insertDateColumn, 'DATETIME2(6)'])
        
        self.__tableColumns = inputTableColumns
        
        # create warehouse table if not exists        
        columnDefs = map(lambda x: f'{x[0]} {x[1]} NULL', inputTableColumns)
        column_defs_join = ',\n'.join(columnDefs)
        createTableSql = f'''
CREATE TABLE [{self.__targetSchemaName}].[{self.__targetTableName}](
{column_defs_join}
)
'''
        if not self.__tableExists:
            self.__execute_dml(createTableSql)
            self.__insertNARecord()
            self.__append_target()
            drop_temp_sql = f'DROP TABLE IF EXISTS `{self.__destinationLakehouse}`.`{self.__tempTableName}`'
            self._spark.sql(drop_temp_sql)
            self.__odbcConnection.close()
            self._completed = True
            return

        # sync schema if table exists
        origColumnsTypesDf = self.__get_current_table_structure()

        consistent_changes = True
        schema_change = False

        if len(origColumnsTypesDf) < len(inputTableColumns):
            schema_change = True

        broken_column_name = None
        broken_column_type = None
        type_error = None

        # find orig columns in new columns (added columns are ok)
        for orig_col in origColumnsTypesDf:
            orig_col_name = orig_col.column_name
            orig_col_type = orig_col.type_name.upper()
            for input_col in inputTableColumns:
                match_found = False
                if input_col[0] == orig_col_name:
                    match_found = True
                    input_type = input_col[1]
                    # type change - nogo
                    input_type_base = input_type.replace('(6)', '').replace('(8000)', '')
                    if input_type_base != orig_col_type:
                        consistent_changes = False
                        schema_change = True
                        broken_column_name = orig_col_name
                        broken_column_type = orig_col_type
                        type_error = f'type does not match source type {input_type}'
                        break
                    # decimal - check the precision and scale
                    if orig_col_type == 'DECIMAL':
                        orig_precision = orig_col.precision
                        orig_scale = orig_col.scale
                        input_decimal_mid = input_type.replace('DECIMAL(', '').replace(')', '').split(',')
                        input_decimal_precision = int(input_decimal_mid[0])
                        input_decimal_scale = int(input_decimal_mid[1])
                        if orig_precision != input_decimal_precision or orig_scale != input_decimal_scale:
                            consistent_changes = False
                            schema_change = True
                            broken_column_name = orig_col_name
                            broken_column_type = orig_col_type
                            type_error = f'precision or scale ({orig_precision},{orig_scale}) does not match source type {input_type}'
                            break
                    # no type difference found - go to next orig table column
                    break
            if not match_found:
                schema_change = True
                consistent_changes = False
                broken_column_name = orig_col_name
                broken_column_type = orig_col_type
                type_error = 'column deleted'
                # deleted or renamed column - nogo
                break
            if consistent_changes == False:
                break

        if consistent_changes == False:
            # problematic schema changes, but it's overwrite, so just drop and create the table
            if incrementMethod == 'overwrite':
                dropTableSql = f'''
DROP TABLE [{self.__targetSchemaName}].[{self.__targetTableName}]
'''
                lgr.info('Recreating the table due to schema changes')
                self.__execute_dml(dropTableSql)
                self.__execute_dml(createTableSql)
                self.__append_target()
                self.__odbcConnection.close()
                self._completed = True
                return
            
            raise Exception(f'Cannot resolve schema chnages in table [{self.__targetSchemaName}].[{self.__targetTableName}], column {broken_column_name}({broken_column_type}) - {type_error}')
            
        if schema_change:
            canAddColumns = self.__tableConfig.canAddNewColumns
            if not canAddColumns:
                raise Exception(f'Cannot resolve schema chnages in destination table [{self.__targetSchemaName}].[{self.__targetTableName}] because the `canAddNewColumns` setting is False')
            
            msg = f'Rebuilding table [{self.__targetSchemaName}].[{self.__targetTableName}], to add new columns...'
            self.__logger.info(msg)
            print(msg)

            # create temp_old copy of the old table in the WH
            # copy data to the temp_old table
            # SELECT * INTO Dim_Customer_temp FROM Dim_Customer
            whTempTableName = f"temp_{self.__targetTableName}_{str(uuid.uuid4())}".replace('-', '_')
            selectIntoQuery = f"SELECT * INTO [{self.__targetSchemaName}].[{whTempTableName}] FROM [{self.__targetSchemaName}].[{self.__targetTableName}]"
            self.__execute_dml(selectIntoQuery)
            
            
            # DROP old table
            dropQuery = f"DROP TABLE [{self.__targetSchemaName}].[{self.__targetTableName}]"
            self.__execute_dml(dropQuery)

            # create table with new schema
            self.__execute_dml(createTableSql)

            # copy data from temp_old table to new table
            insert_columns = list(map(lambda x: f'[{x.column_name}]', origColumnsTypesDf))
            insert_columns_join = ", \n".join(insert_columns)
            insertBackQuery = f'''INSERT INTO [{self.__targetSchemaName}].[{self.__targetTableName}]
            ({insert_columns_join})
            SELECT {insert_columns_join}
            FROM [{self.__targetSchemaName}].[{whTempTableName}]
            '''
            self.__execute_dml(insertBackQuery)

            # DROP temp_old table
            dropQuery = f'DROP TABLE [{self.__targetSchemaName}].[{whTempTableName}]'
            self.__execute_dml(dropQuery)
        
        if incrementMethod != 'overwrite':
            # overwrite will handle it's N/A record by itself
            self.__insertNARecord()

        # handle increments as in a lakehouse
        if incrementMethod == 'overwrite':
            self.__overwrite_target()
        elif incrementMethod == 'append':
            self.__append_target()
        elif incrementMethod == 'merge':
            self.__merge_taget()
        elif incrementMethod == 'snapshot':
            self.__snapshot_target()
        else:
            raise Exception(f'Unrecognized increment type: {incrementMethod}')

        self.__odbcConnection.close()
        drop_temp_sql = f'DROP TABLE IF EXISTS `{self.__destinationLakehouse}`.`{self.__tempTableName}`'
        self._spark.sql(drop_temp_sql)

        self._completed = True


    def __execute_select(self, query: str):
        pd_df = pd.read_sql(query, self.__odbcConnection)
        df = self._spark.createDataFrame(pd_df)
        return df.collect()
    
    def __execute_dml(self, query: str):
        queryOneLine = query.replace('\n', ' ')
        self.__logger.info(queryOneLine)
        try:
            self.__odbcConnection.execute(query)
            self.__odbcConnection.commit()
        except Exception as e:
            if str(e.args[0]) == '42000' and e.args[1].find('the object accessed by the statement has been modified by a DDL statement in another concurrent transaction') > -1:
                warn = 'Waiting 5 s for blocking DDL to finish'
                print(warn)
                self.__logger.warning(e.args[1])
                self.__logger.warning(warn)
                time.sleep(5)
                self.__odbcConnection.execute(query)
                self.__odbcConnection.commit()
            elif str(e.args[0]) == '42S02' and e.args[1].find('Invalid object name') > -1 and e.args[1].find('temp') > -1:
                warn = 'Waiting 5 s for the lakehouse temp table to come online'
                print(warn)
                self.__logger.warning(e.args[1])
                self.__logger.warning(warn)
                time.sleep(5)
                try:
                    self.__odbcConnection.execute(query)
                    self.__odbcConnection.commit()
                except Exception as ee:
                    if str(ee.args[0]) == '42000' and ee.args[1].find('the object accessed by the statement has been modified by a DDL statement in another concurrent transaction') > -1:
                        warn = 'Waiting 5 s for blocking DDL to finish'
                        print(warn)
                        self.__logger.warning(ee.args[1])
                        self.__logger.warning(warn)
                        time.sleep(5)
                        self.__odbcConnection.execute(query)
                        self.__odbcConnection.commit()
                    else:
                        raise ee
            else:
                raise e

    def __get_current_table_structure(self):
        df = self.__execute_select(f'''
SELECT s.name schema_name, t.name table_name, c.name column_name, tt.name type_name, tt.max_length, tt.precision, tt.scale 
FROM sys.schemas s
INNER JOIN sys.tables t ON s.schema_id = t.schema_id
INNER JOIN sys.columns c ON c.object_id = t.object_id
INNER JOIN sys.types tt ON tt.system_type_id = c.system_type_id
WHERE s.name = '{self.__targetSchemaName}' AND t.name = '{self.__targetTableName}'
        ''')
        return df
    
    def __list_diff(self, first_list, second_list):
        diff = [item for item in first_list if item not in second_list]
        return diff
    
    # INSERT INTO destination_table(destination_columns) SELECT {destination_columns, including TS and ID} FROM src
    # [src] to be defined later
    def __create_insert_query(self):
        col_list = list(map(lambda x: x[0], self.__tableColumns))
        insert_sql = ''
        select_sql = ''
        for i in range(len(col_list)):
            if i > 0:
                insert_sql = insert_sql + ',\n'
                select_sql = select_sql + ',\n'
            col_name = col_list[i]
            insert_sql = insert_sql + f'[{col_name}]'
            
            if col_name == self.__identityColumn:
                init_id = 0
                if self.__incrementMethod != 'overwrite' and self.__tableExists:
                    maxIdDf = self.__execute_select(f'SELECT ISNULL(MAX([{self.__identityColumn}]), 0) AS MaxID FROM [{self.__targetSchemaName}].[{self.__targetTableName}] ')
                    init_id = maxIdDf[0][0]
                select_sql = select_sql + f'ROW_NUMBER() OVER(ORDER BY (SELECT NULL)) + {init_id} [{col_name}]'
            elif col_name == self.__insertDateColumn:
                select_sql = select_sql + f'GETDATE() [{col_name}]'
            else:
                select_sql = select_sql + f'src.[{col_name}]'
            
        complete_sql = f'''INSERT INTO [{self.__targetSchemaName}].[{self.__targetTableName}](
{insert_sql}        
)
SELECT
{select_sql}
FROM src
'''
        return complete_sql
    
    def __append_target_query(self, src_query):
        insert_query = self.__create_insert_query()
        append_sql = f'''WITH
src AS ({src_query}
)
{insert_query}
'''
        self.__execute_dml(append_sql)

    def __insertNARecord(self):        
        self.__addNARecord = self.__tableConfig.addNARecord
        if not self.__addNARecord:
            return
        if self.__identityColumn is None:
            raise Exception('Configuration error - when addNARecord is enabled, identityColumnPattern needs to be configured as well.')
                
        # IF NOT EXISTS(SELECT TOP 1 1 FROM [dbo].[SurveyData] WHERE [SurveyDataID] = -1)
        # BEGIN
        #     INSERT INTO [dbo].[SurveyData] ([SurveyDataID], [Year], [Industry_aggregation_NZSIOC])
        #     VALUES(-1, 0, 'N/A')
        # END

        columnsTypesDf = self.__get_current_table_structure()

        insert_list = []
        value_list = []

        for c in columnsTypesDf:
            col_name = c.column_name
            col_type = c.type_name.upper()
            insert_list.append(col_name)
            if col_name == self.__identityColumn:
                value_list.append('-1')
            elif col_type in ['INT', 'BIGINT', 'SMALLINT', 'BIT']:
                value_list.append('0')
            elif col_type in ['REAL', 'FLOAT']:
                value_list.append('0.0')
            elif col_type in ['DATETIME', 'DATETIME2']:
                value_list.append("'2000-01-01'")
            elif col_type in ['VARCHAR', 'CHAR']:
                value_list.append("'N/A'")
        
        insert_columns = list(map(lambda x: f'[{x}]', insert_list))
        insert_columns_join = ", \n".join(insert_columns)
        values_join = ", \n".join(value_list)

        insert_query = f'''
IF NOT EXISTS(SELECT TOP 1 1 FROM [{self.__targetSchemaName}].[{self.__targetTableName}] WHERE [{self.__identityColumn}] = -1)
BEGIN
    INSERT INTO [{self.__targetSchemaName}].[{self.__targetTableName}] ({insert_columns_join})
    VALUES({values_join})
END
'''
        self.__execute_dml(insert_query)
    
    def __append_target(self):
        src_query = f'SELECT * FROM [{self.__destinationLakehouse}].[dbo].[{self.__tempTableName}]'
        self.__append_target_query(src_query)
    
    def __overwrite_target(self):
        delte_sql = f'DELETE FROM [{self.__targetSchemaName}].[{self.__targetTableName}]'
        self.__execute_dml(delte_sql)
        self.__insertNARecord()
        self.__append_target()
    
    def __merge_taget(self):
        all_columns = list(map(lambda x: x[0], self.__tableColumns))
        key_columns = self.__tableConfig.mergeKeyColumns
        non_key_columns = self.__list_diff(self.__list_diff(all_columns, key_columns), [self.__identityColumn, self.__insertDateColumn])
        
        # print('key columns')
        # print(key_columns)
        # print('non-key columns')
        # print(non_key_columns)
        
        if len(key_columns) == 0:
            raise Exception('No key columns set for merge increment. Please set the mergeKeyColumns property in destinationTable configuration to the list of column names.')
        
        join_condition = " AND ".join([f"src.[{item}] = tgt.[{item}]" for item in key_columns])
        
        if len(non_key_columns) > 0:
            update_list = ",\n".join([f"tgt.[{item}] = src.[{item}]" for item in non_key_columns])
            update_query = f'''
UPDATE tgt SET
{update_list}
FROM [{self.__destinationLakehouse}].[dbo].[{self.__tempTableName}] src
INNER JOIN [{self.__targetSchemaName}].[{self.__targetTableName}] tgt ON {join_condition}
'''
            self.__execute_dml(update_query)

        insert_src_query = f'''
SELECT src.*
FROM [{self.__destinationLakehouse}].[dbo].[{self.__tempTableName}] src
LEFT JOIN [{self.__targetSchemaName}].[{self.__targetTableName}] tgt ON {join_condition}
WHERE tgt.{key_columns[0]} IS NULL
'''
        self.__append_target_query(insert_src_query)

    def __snapshot_target(self):
        # first delete the snapshot to be replaced
        key_columns = list(map(lambda x: self.__sanitizeColumnName(x), self.__tableConfig.snapshotKeyColumns))
        
        if len(key_columns) == 0:
            raise Exception('No key columns set for snapshot increment. Please set the snapshotKeyColumns property in destinationTable configuration to the list of column names.')
        
        join_condition = " AND ".join([f"src.[{item}] = tgt.[{item}]" for item in key_columns])
        delete_query = f'''
DELETE tgt FROM [{self.__targetSchemaName}].[{self.__targetTableName}] tgt
INNER JOIN [{self.__destinationLakehouse}].[dbo].[{self.__tempTableName}] src ON {join_condition}
'''
        self.__execute_dml(delete_query)
        self.__append_target()

        
    def __filterByWatermark(self):
        if not(self.__tableExists):
            return
        elif self.__incrementMethod is None:
            return
        elif self.__incrementMethod == 'overwrite':
            return
        
        watermarkColumn = self.__tableConfig.watermarkColumn
        if watermarkColumn is None:
            return
        

        ###############
        # DDL conflict would occur when using watermark before introducing the 5s retry in __execute_dml
        #
        # pyodbc.ProgrammingError: ('42000', "[42000] [Microsoft][ODBC Driver 18 for SQL Server][SQL Server]Snapshot isolation transaction failed in database 'LH_exp' because the object accessed by the statement has been modified by a DDL statement in another concurrent transaction since the start of this transaction.  It is disallowed because the metadata is not versioned. A concurrent update to metadata can lead to inconsistency if mixed with snapshot isolation. (3961) (SQLExecDirectW)")
        #
        #     361         insert_query = self.__create_insert_query()
        #     362         append_sql = f'''WITH
        #     363 src AS ({src_query}
        #     364 )
        #     365 {insert_query}
        #     366 '''
        # --> 367         self.__execute_dml(append_sql)
        ###############

        self.__watermarkColumn = watermarkColumn
        max_watermark_sql = f'SELECT COALESCE(MAX([{watermarkColumn}]), \'\') FROM [{self.__targetSchemaName}].[{self.__targetTableName}]'
        # print(max_watermark_sql)
        max_watermark = self.__execute_select(max_watermark_sql)[0][0]
        # print(f'max watermark: {max_watermark}')

        # if the table is empty or watermark is 0, take all the data
        if max_watermark is None:
            return
        if not max_watermark:
            return
        
        filter = f'{watermarkColumn} > "{max_watermark}"'
        self.__data = self.__data.filter(filter)
        # print(filter)
        # print(self.__data.count())
    
    def __sanitizeColumnName(self, colName):
        replacement = self.__tableConfig.invalidCharactersInColumnNamesReplacement
        if replacement is None:
            return colName
        
        invalids = " ,;{}()\n\t="
        name = colName        
        for i in invalids:
            name = name.replace(i, replacement)
        return name
    
    def __replaceInvalidCharactersInColumnNames(self):
        replacement = self.__tableConfig.invalidCharactersInColumnNamesReplacement
        if replacement is None:
            return
        
        self.__data = (
            self.
            __data
            .select(
                [col(f'`{c}`').alias(
                    self.__sanitizeColumnName(c)) 
                    for c in self.__data.columns
                ]))


from bifabrik.cfg.engine.Configuration import Configuration
from bifabrik.cfg.engine.Configuration import CfgProperty

class DestinationStorageConfiguration(Configuration):
    """Defines the data destination lakehouse or warehouse. By default refers to the default lakehouse of the notebook.
    """
    def __init__(self):
        self._explicitProps = {}
        self.__destinationWorkspace = None
        self.__destinationLakehouse = None
        self.__destinationWarehouseConnectionString = None
        self.__destinationWarehouseConnectionTimeout = 600
        self.__destinationWarehouseName = None

    @CfgProperty
    def destinationWorkspace(self) -> str:
        """The workspace to which pipeline destinations writes data - ID or name
        """
        return self.__destinationWorkspace
    @destinationWorkspace.setter(key='destinationWorkspace')
    def destinationWorkspace(self, val):
        self.__destinationWorkspace = val

    @CfgProperty
    def destinationLakehouse(self) -> str:
        """The lakehouse to which pipeline destinations writes data - ID or name
        """
        return self.__destinationLakehouse
    @destinationLakehouse.setter(key='destinationLakehouse')
    def destinationLakehouse(self, val):
        self.__destinationLakehouse = val

    @CfgProperty
    def destinationWarehouseConnectionString(self) -> str:
        """Connection string to the Fabric warehouse (server name like dxt....datawarehouse.fabric.microsoft.com)
        """
        return self.__destinationWarehouseConnectionString
    @destinationWarehouseConnectionString.setter(key='destinationWarehouseConnectionString')
    def destinationWarehouseConnectionString(self, val):
        self.__destinationWarehouseConnectionString = val

    @CfgProperty
    def destinationWarehouseConnectionTimeout(self) -> str:
        """Warehouse ODBC connection timetout in seconds
        default 600
        """
        return self.__destinationWarehouseConnectionTimeout
    @destinationWarehouseConnectionTimeout.setter(key='destinationWarehouseConnectionTimeout')
    def destinationWarehouseConnectionTimeout(self, val):
        self.__destinationWarehouseConnectionTimeout = val

    @CfgProperty
    def destinationWarehouseName(self) -> str:
        """Name of the warehouse to which pipeline destinations writes data
        """
        return self.__destinationWarehouseName
    @destinationWarehouseName.setter(key='destinationWarehouseName')
    def destinationWarehouseName(self, val):
        self.__destinationWarehouseName = val


from bifabrik.cfg.engine.ConfigContainer import ConfigContainer
from bifabrik.cfg.specific.JsonConfiguration import JsonConfiguration
from bifabrik.cfg.specific.FileSourceConfiguration import FileSourceConfiguration
from bifabrik.cfg.DataSourceConfigurationBase import DataSourceConfigurationBase

class JsonSourceConfiguration(DataSourceConfigurationBase):

    def __init__(self):
        self.__json = JsonConfiguration()
        self.__fileSource = JsonConfiguration()
        super().__init__()
        

    @property
    def json(self) -> JsonConfiguration:
        return self.__json
    
    @property
    def fileSource(self) -> FileSourceConfiguration:
        return self.__fileSource


from bifabrik.cfg.engine.ConfigContainer import ConfigContainer
from bifabrik.cfg.specific.DestinationStorageConfiguration import DestinationStorageConfiguration

class DataDestinationConfigurationBase(ConfigContainer):

    def __init__(self):
        self.__destinationStorage = DestinationStorageConfiguration()
        super().__init__()
        

    @property
    def destinationStorage(self) -> DestinationStorageConfiguration:
        """The destination lakehouse and warehouse reference used by pipelines"""
        return self.__destinationStorage


from bifabrik.cfg.engine.Configuration import Configuration
from bifabrik.cfg.engine.Configuration import CfgProperty

class LogConfiguration(Configuration):
    """Logging configuration
    """
    def __init__(self):
        self._explicitProps = {}
        self.__loggingEnabled = True
        self.__logLakehouse = None
        self.__logWorkspace = None
        self.__loggingLevel = 'INFO'
        self.__errorLogPath = 'Files/BifabrikErrorLog.log'
        self.__logPath = 'Files/BifabrikLog.log'
        self.__modified = False

    @CfgProperty
    def loggingEnabled(self) -> bool:
        """True / False whether logging is enabled
        """
        return self.__loggingEnabled
    @loggingEnabled.setter(key='loggingEnabled')
    def loggingEnabled(self, val):
        if self.__loggingEnabled != val:
            self.__modified = True
        self.__loggingEnabled = val
        

    # @CfgProperty
    # def logLakehouse(self) -> str:
    #     """The lakehouse to which logs will be saved
    #     """
    #     return self.__logLakehouse
    # @logLakehouse.setter(key='logLakehouse')
    # def logLakehouse(self, val):
    #     if self.__logLakehouse != val:
    #         self.__modified = True
    #     self.__logLakehouse = val

    @CfgProperty
    def loggingLevel(self) -> str:
        """DEBUG / INFO / WARNING / ERROR / CRITICAL
        """
        return self.__loggingLevel
    @loggingLevel.setter(key='loggingLevel')
    def loggingLevel(self, val):
        if self.__loggingLevel != val:
            self.__modified = True
        self.__loggingLevel = val

    @CfgProperty
    def errorLogPath(self) -> str:
        """The file to which to save error logs
        """
        return self.__errorLogPath
    @errorLogPath.setter(key='errorLogPath')
    def errorLogPath(self, val):
        if self.__errorLogPath != val:
            self.__modified = True
        self.__errorLogPath = val
        
    @CfgProperty
    def logPath(self) -> str:
        """The file to which to save logs
        """
        return self.__logPath
    @logPath.setter(key='logPath')
    def logPath(self, val):
        if self.__logPath != val:
            self.__modified = True
        self.__logPath = val

    @property
    def modified(self) -> bool:
        return self.__modified
    @modified.setter
    def modified(self, val) -> bool:
        self.__modified = val

    # logging to a different lakehouse cannot be supported now, because the default logger aims at a mounted directory in the environment
    # and other lakeouses are not in mounts
    
    # @CfgProperty
    # def logWorkspace(self) -> str:
    #     """The workspace (id or name) to which to save logs
    #     """
    #     return self.__logWorkspace
    # @logWorkspace.setter(key='logWorkspace')
    # def logWorkspace(self, val):
    #     if self.__logWorkspace != val:
    #         self.__modified = True
    #     self.__logWorkspace = val
    
    # @CfgProperty
    # def logLakehouse(self) -> str:
    #     """The lakehouse (id or name) to which to save logs
    #     """
    #     return self.__logLakehouse
    # @logLakehouse.setter(key='logLakehouse')
    # def logLakehouse(self, val):
    #     if self.__logLakehouse != val:
    #         self.__modified = True
    #     self.__logLakehouse = val


from bifabrik.cfg.engine.Configuration import Configuration
from bifabrik.cfg.engine.Configuration import CfgProperty

class SecurityConfiguration(Configuration):
    """Credentials configuration, currently used for connecting to Fabric warehouses and the SharePointListSource. (Passwords aren't actually stored here, this configuration only refers to KeyVault secrets.)
    """
    def __init__(self):
        self._explicitProps = {}
        self.__keyVaultUrl = None
        self.__loginKVSecretName = None
        self.__passwordKVSecretName = None
        self.__servicePrincipalClientId = None
        self.__servicePrincipalClientSecretKVSecretName = None

    @CfgProperty
    def keyVaultUrl(self) -> str:
        """Key Vault URL - e.g. 'https://kv-fabrik.vault.azure.net/'
        """
        return self.__keyVaultUrl
    @keyVaultUrl.setter(key='keyVaultUrl')
    def keyVaultUrl(self, val):
        self.__keyVaultUrl = val

    @CfgProperty
    def loginKVSecretName(self) -> str:
        """Secret in the Key Vault that contains the username
        """
        return self.__loginKVSecretName
    @loginKVSecretName.setter(key='loginKVSecretName')
    def loginKVSecretName(self, val):
        self.__loginKVSecretName = val
    
    @CfgProperty
    def passwordKVSecretName(self) -> str:
        """Secret in the Key Vault that contains the password
        """
        return self.__passwordKVSecretName
    @passwordKVSecretName.setter(key='passwordKVSecretName')
    def passwordKVSecretName(self, val):
        self.__passwordKVSecretName = val

    @CfgProperty
    def servicePrincipalClientId(self) -> str:
        """Service principal used for connecting to Fabric warehouses
        """
        return self.__servicePrincipalClientId
    @servicePrincipalClientId.setter(key='servicePrincipalClientId')
    def servicePrincipalClientId(self, val):
        self.__servicePrincipalClientId = val
    
    @CfgProperty
    def servicePrincipalClientSecretKVSecretName(self) -> str:
        """Secret in the Key Vault that contains the service principal's client secret
        """
        return self.__servicePrincipalClientSecretKVSecretName
    @servicePrincipalClientSecretKVSecretName.setter(key='servicePrincipalClientSecretKVSecretName')
    def servicePrincipalClientSecretKVSecretName(self, val):
        self.__servicePrincipalClientSecretKVSecretName = val

    


#from pyspark.sql.session import SparkSession
from bifabrik.dst.LakehouseTableDestination import LakehouseTableDestination
from bifabrik.dst.WarehouseTableDestination import WarehouseTableDestination
from bifabrik.dst.SparkDfDestination import SparkDfDestination
from bifabrik.dst.PandasDfDestination import PandasDfDestination
from bifabrik.base.Task import Task
#from typing import Any, Self

class DataTransformation(Task):
    """Generic data source. 
    Contains methods to initiate a DataDestination after you are done setting the transformation.
    bif.sourceInit.sourceOption1(A).srcOpt2(B).transformation1(X).transformation2(Y).toTable("Tab").destinationOption(X)....run()
    """
    def __init__(self, parentPipeline):
        super().__init__(parentPipeline)
        
    def toTable(self, targetTableName: str) -> LakehouseTableDestination:
        """Sets the lakehouse destination table name
        """
        return self.toLakehouseTable(targetTableName)
    
    def toLakehouseTable(self, targetTableName: str) -> LakehouseTableDestination:
        """Sets the lakehouse destination table name
        """
        dst = LakehouseTableDestination(self._pipeline, targetTableName)
        return dst
    
    def toWarehouseTable(self, targetTableName: str, targetSchemaName: str = 'dbo') -> WarehouseTableDestination:
        """Sets the warehouse destination table name. The warehouse authentication needs to be configured - see https://rjankovic.github.io/bifabrik/tutorial/dst_warehouse_table.html

        Examples
        --------
        > import bifabrik as bif
        >
        > bif.config.security.servicePrincipalClientId = '56712345-1234-7890-abcd-abcd12344d14'
        > bif.config.security.keyVaultUrl = 'https://kv-contoso.vault.azure.net/'
        > bif.config.security.servicePrincipalClientSecretKVSecretName = 'contoso-clientSecret'
        > bif.config.destinationStorage.destinationWarehouseName = 'WH_GOLD'
        > bif.config.destinationStorage.destinationWarehouseConnectionString = 'dxtxxxxxxbue.datawarehouse.fabric.microsoft.com'
        >
        > bif.fromSql('''
        > SELECT countryOrRegion `CountryOrRegion`
        > ,YEAR(date) `Year` 
        > ,COUNT(*) `PublicHolidayCount`
        > FROM LH_SILVER.publicholidays
        > GROUP BY countryOrRegion
        > ,YEAR(date)
        > ''').toWarehouseTable('HolidayCountsYearly').run()
        """
        dst = WarehouseTableDestination(self._pipeline, targetTableName, targetSchemaName)
        return dst
    
    def toSparkDf(self) -> SparkDfDestination:
        """Sets the destination as a Spark DF (for further processing)
        """
        dst = SparkDfDestination(self._pipeline)
        return dst

    def toPandasDf(self) -> PandasDfDestination:
        """Sets the destination as a Pandas DF (for further processing)
        """
        dst = PandasDfDestination(self._pipeline)
        return dst
    
    def transformSparkDf(self, func):
        """Applies the givet function to transform the data as a Spark DF
        """
        from .SparkDfTransformation import SparkDfTransformation
        tsf = SparkDfTransformation(self._pipeline, func)
        return tsf
    
    def transformPandasDf(self, func):
        """Applies the givet function to transform the data as a Pandas DF
        """
        from .PandasDfTransformation import PandasDfTransformation
        tsf = PandasDfTransformation(self._pipeline, func)
        return tsf
        
    def validate(self, testName = ''):
        """Test the input data by checking the ValidationResult and ValidationMessage columns to determine if each row failed / passed the validation
        """
        from .ValidationTransformation import ValidationTransformation
        tsf = ValidationTransformation(self._pipeline, testName)
        return tsf


#from pyspark.sql.session import SparkSession
from pyspark.sql.dataframe import DataFrame
from bifabrik.dst.LakehouseTableDestination import LakehouseTableDestination
from bifabrik.dst.WarehouseTableDestination import WarehouseTableDestination
from bifabrik.dst.SparkDfDestination import SparkDfDestination
from bifabrik.dst.PandasDfDestination import PandasDfDestination
from bifabrik.tsf.DataTransformation import DataTransformation
from bifabrik.tsf import SparkDfTransformation
from bifabrik.tsf import PandasDfTransformation
from bifabrik.tsf import ValidationTransformation
from bifabrik.base.Task import Task

class DataSource(Task):
    """Generic data source. 
    Contains methods to initiate a DataDestination after you are done setting the source.
    bif.sourceInit.sourceOption1(A).srcOpt2(B).toTable("Tab").destinationOption(X)....save()
    """
    def __init__(self, parentPipeline):
        super().__init__(parentPipeline)
        

    # def toDf(self) -> DataFrame:
    #     """Converts the data to a spark dataframe
    #     """
    #     pass

    def toTable(self, targetTableName: str) -> LakehouseTableDestination:
        """Sets the lakehouse destination table name
        """
        return self.toLakehouseTable(targetTableName)
    
    def toLakehouseTable(self, targetTableName: str) -> LakehouseTableDestination:
        """Sets the lakehouse destination table name
        """
        dst = LakehouseTableDestination(self._pipeline, targetTableName)
        return dst
    
    def toWarehouseTable(self, targetTableName: str, targetSchemaName: str = 'dbo') -> WarehouseTableDestination:
        """Sets the warehouse destination table name. The warehouse authentication needs to be configured - see https://rjankovic.github.io/bifabrik/tutorial/dst_warehouse_table.html

        Examples
        --------
        > import bifabrik as bif
        >
        > bif.config.security.servicePrincipalClientId = '56712345-1234-7890-abcd-abcd12344d14'
        > bif.config.security.keyVaultUrl = 'https://kv-contoso.vault.azure.net/'
        > bif.config.security.servicePrincipalClientSecretKVSecretName = 'contoso-clientSecret'
        > bif.config.destinationStorage.destinationWarehouseName = 'WH_GOLD'
        > bif.config.destinationStorage.destinationWarehouseConnectionString = 'dxtxxxxxxbue.datawarehouse.fabric.microsoft.com'
        >
        > bif.fromSql('''
        > SELECT countryOrRegion `CountryOrRegion`
        > ,YEAR(date) `Year` 
        > ,COUNT(*) `PublicHolidayCount`
        > FROM LH_SILVER.publicholidays
        > GROUP BY countryOrRegion
        > ,YEAR(date)
        > ''').toWarehouseTable('HolidayCountsYearly').run()
        """
        dst = WarehouseTableDestination(self._pipeline, targetTableName, targetSchemaName)
        return dst
    
    def toSparkDf(self) -> SparkDfDestination:
        """Sets the destination as a Spark DF (for further processing)
        """
        dst = SparkDfDestination(self._pipeline)
        return dst
    
    def toPandasDf(self) -> PandasDfDestination:
        """Sets the destination as a Pandas DF (for further processing)
        """
        dst = PandasDfDestination(self._pipeline)
        return dst
    
    def transformSparkDf(self, func) -> SparkDfTransformation.SparkDfTransformation:
        """Applies the givet function to transform the data as a Spark DF
        """
        tsf = SparkDfTransformation.SparkDfTransformation(self._pipeline, func)
        return tsf
    
    def transformPandasDf(self, func) -> PandasDfTransformation.PandasDfTransformation:
        """Applies the givet function to transform the data as a Pandas DF
        """
        tsf = PandasDfTransformation.PandasDfTransformation(self._pipeline, func)
        return tsf
    
    def validate(self, testName = '') -> ValidationTransformation.ValidationTransformation:
        """Test the input data by checking the ValidationResult and ValidationMessage columns to determine if each row failed / passed the validation.
        """
        tsf = ValidationTransformation.ValidationTransformation(self._pipeline, testName)
        return tsf


from bifabrik.cfg.engine.ConfigContainer import ConfigContainer
from bifabrik.cfg.specific.ValidationConfiguration import ValidationConfiguration

class ValidationTransformationConfiguration(ConfigContainer):

    def __init__(self):
        self.__validation = ValidationConfiguration()
        super().__init__()
        

    @property
    def validation(self) -> ValidationConfiguration:
        """Data validation configuration"""
        return self.__validation


from bifabrik.cfg.engine.Configuration import Configuration
from bifabrik.cfg.engine.Configuration import CfgProperty

class TableConfiguration(Configuration):
    """Configuration related to saving data to delta tables in a Fabric lakehouse or warehouse
    """
    def __init__(self):
        self._explicitProps = {}
        self.__watermarkColumn = None
        self.__increment = None
        self.__identityColumnPattern = None
        self.__insertDateColumn = None
        self.__mergeKeyColumns = []
        self.__snapshotKeyColumns = []
        self.__invalidCharactersInColumnNamesReplacement = '_'
        self.__canAddNewColumns = True
        self.__addNARecord = False

    @CfgProperty
    def watermarkColumn(self) -> str:
        """Which column to use as the watermark when filtering records for incremental load.
        If the target table already exists and the watermark column is set, only records where the watermark is greater than the MAX from the table will be inserted.
        default None
        """
        return self.__watermarkColumn
    @watermarkColumn.setter(key='watermarkColumn')
    def watermarkColumn(self, val):
        self.__watermarkColumn = val

    @CfgProperty
    def increment(self) -> str:
        """Increment method
        None or "overwrite" : full load
        "append"            : add incoming rows to the table
        "merge"             : update / insert (SCD 1) after matching based on the mergeKeyColumns setting
        "snapshot"          : based on the snapshotKeyColumns setting (typically a date), inseert the new records and remove any old records that refer to a newly loaded snapshot key
        default None
        """
        return self.__increment
    @increment.setter(key='increment')
    def increment(self, val):
        self.__increment = val

    @CfgProperty
    def mergeKeyColumns(self) -> str:
        """Array of column names used as the business key columns when the 'merge' increment method is used
        default []
        """
        return self.__mergeKeyColumns
    @mergeKeyColumns.setter(key='mergeKeyColumns')
    def mergeKeyColumns(self, val):
        self.__mergeKeyColumns = val

    @CfgProperty
    def snapshotKeyColumns(self) -> str:
        """Array of column names used as the snapshot key columns when the 'snapshot' increment method is used
        default []
        """
        return self.__snapshotKeyColumns
    @snapshotKeyColumns.setter(key='snapshotKeyColumns')
    def snapshotKeyColumns(self, val):
        self.__snapshotKeyColumns = val

    @CfgProperty
    def identityColumnPattern(self) -> str:
        """The name pattern of the identity column to be added to the table. The values are 1, 2, 3..., each new record gets a new number.
        If configured, the column will be placed at the beginning of the table.
        
        Supported placeholders:
        {tablename} : the delta table name
        {lakehousename} : the lakehouse name

        example: '{tablename}_ID'

        default None
        """
        return self.__identityColumnPattern
    @identityColumnPattern.setter(key='identityColumnPattern')
    def identityColumnPattern(self, val):
        self.__identityColumnPattern = val

    @CfgProperty
    def insertDateColumn(self) -> str:
        """The name of the timestamp column to be added to each record, reflecting the time when the record was inserted.
        default None
        """
        return self.__insertDateColumn
    @insertDateColumn.setter(key='insertDateColumn')
    def insertDateColumn(self, val):
        self.__insertDateColumn = val

    @CfgProperty
    def invalidCharactersInColumnNamesReplacement(self) -> str:
        """The characters ' ,;{}()\n\t=' are invalid in delta tables. This sets the replacement character. You can use '' to just remove the invalid characters.
        default '_'
        """
        return self.__invalidCharactersInColumnNamesReplacement
    @invalidCharactersInColumnNamesReplacement.setter(key='invalidCharactersInColumnNamesReplacement')
    def invalidCharactersInColumnNamesReplacement(self, val):
        self.__invalidCharactersInColumnNamesReplacement = val
    
    @CfgProperty
    def canAddNewColumns(self) -> bool:
        """If the table already exists and the input dataset has extra columns, add the columns to the table with empty values before loading the new input.
        default True
        """
        return self.__canAddNewColumns
    @canAddNewColumns.setter(key='canAddNewColumns')
    def canAddNewColumns(self, val):
        self.__canAddNewColumns = val
    
    @CfgProperty
    def addNARecord(self) -> bool:
        """Add a dimension "N/A" record with identity -1, N/A for strings, 0 for numbers. identityColumnPattern needs to be configured when this is enabled.
        default False
        """
        return self.__addNARecord
    @addNARecord.setter(key='addNARecord')
    def addNARecord(self, val):
        self.__addNARecord = val



"""The starting point for using the library.

    For more info see https://rjankovic.github.io/bifabrik/
    
    To create an instance, pass the :class:`SparkSession` ("spark") from your Fabric notebook.
    You can then use various methods to load data to tables to the current lakehouse.
    The notebook has to have a default lakehouse selected - it references files from the Files/
    path of the lakehouse and tables from Tables/

    Examples
    --------
    Load a CSV file from Files/Folder/orders*.csv (all files matching the pattern) to a table called OrdersTable.

    >>> import bifabrik as bif
    >>>
    >>> bif.fromCsv.path('Folder/orders*.csv').toTable('OrdersTable').run()
    
    Load the results of a SQL query to a table

    >>> bif.fromSql("SELECT * FROM OrdersTable LIMIT 10").toTable('TenOrders').save()
"""

from bifabrik.src.CsvSource import CsvSource
from bifabrik.src.ExcelSource import ExcelSource
from bifabrik.src.SharePointListSource import SharePointListSource
from bifabrik.src.JsonSource import JsonSource
from bifabrik.src.SparkSqlSource import SparkSqlSource
from bifabrik.src.WarehouseSqlSource import WarehouseSqlSource
from bifabrik.src.SparkDfSource import SparkDfSource
from bifabrik.src.PandasDfSource import PandasDfSource
from bifabrik.cfg.CompleteConfiguration import CompleteConfiguration
import bifabrik.utils.log as log
from bifabrik.base.Pipeline import Pipeline
import pyspark.sql.session as pss
from pyspark.sql.dataframe import DataFrame as SparkDf
from pandas.core.frame import DataFrame as PandasDf

# read version from installed package
from importlib.metadata import version
__version__ = version("bifabrik")
__loggername__ = "bifabrik_logger"

#import bifabrik
from bifabrik.utils.fsUtils import getDefaultLakehouseAbfsPath

if getDefaultLakehouseAbfsPath() is None:
    print('bifabrik warning: the notebook is not attached to a lakehouse - some features will not work correctly.')


#from bifabrik.base.Task import Task

__spark = pss.SparkSession.builder.getOrCreate()
__configuration = CompleteConfiguration()
        #self.__configuration.log.loggingEnabled = False
    
def __prepPipeline() -> Pipeline:
    global __configuration
    global __spark
    log.configureLogger(__configuration.log)
    return Pipeline(__spark, __configuration)


def fromCsv(path: str = None) -> CsvSource:
    """Load data from CSV
    
    Examples
    --------
    
    >>> import bifabrik as bif
    >>>
    >>> bif.fromCsv('orders*.csv').toTable('Orders').run()
    """
    ds = CsvSource(__prepPipeline())
    ds.path(path)
    return ds


def fromJson(path: str = None) -> JsonSource:
    """Load data from JSON
    
    Examples
    --------
    
    >>> import bifabrik as bif
    >>>
    >>> bif.fromJson('invoices.json').toTable('Invoices').run()
    """
    ds = JsonSource(__prepPipeline())
    ds.path(path)
    return ds

def fromExcel(path: str = None) -> ExcelSource:
    """Load data from Excel
    
    Examples
    --------
    
    >>> import bifabrik as bif
    >>>
    >>> bif.fromExcel.path('ExcelData/factOrderLine.xlsx').sheetName('Sheet1').toTable('FactOrderLine').run()
    """
    ds = ExcelSource(__prepPipeline())
    ds.path(path)
    return ds


def fromSql(query: str = None) -> SparkSqlSource:
    """Load the result of a SparkSQL query to a table
    
    Examples
    --------
    
    >>> import bifabrik as bif
    >>>
    >>> bif.fromSql('SELECT A, B, C FROM Table1 WHERE D = 1').toTable('Table2').run()
    """
    return fromSparkSql(query)

def fromSparkSql(query: str = None) -> SparkSqlSource:
    """Load the result of a SparkSQL query to a table
    
    Examples
    --------
    
    >>> import bifabrik as bif
    >>>
    >>> bif.fromSql('SELECT A, B, C FROM Table1 WHERE D = 1').toTable('Table2').run()
    """
    ds = SparkSqlSource(__prepPipeline())
    ds.query(query)
    return ds

def fromWarehouseSql(query: str) -> WarehouseSqlSource:
    """Load the result of a TSQL query from a Fabric warehouse to a table
    Service principal authentication is used to connect to the warehouse. This needs to be configured as below
    
    Examples
    --------
    >>> import bifabrik as bif
    >>>
    >>> bif.config.security.keyVaultUrl = 'https://kv-contoso.vault.azure.net/'
    >>> bif.config.security.servicePrincipalClientId = '56712345-1234-7890-abcd-abcd12344d14'
    >>> bif.config.security.servicePrincipalClientSecretKVSecretName = 'contoso-clientSecret'
    >>> bif.config.sourceStorage.sourceWarehouseConnectionString = 'dxtxxxxxxbue.datawarehouse.fabric.microsoft.com'
    >>>
    >>> bif.fromWarehouseSql('''
    >>> SELECT CountryOrRegion, Year, PublicHolidayCount FROM [DW1].[dbo].[HolidayCountsYearly]
    >>> ''').toTable('HolidayCountsYearlyFromDW').run()
    
    """
    ds = WarehouseSqlSource(__prepPipeline(), query)
    return ds

def fromSparkDf(df: SparkDf) -> SparkDfSource:
    """Use spark dataframe as source
    
    Examples
    --------
    
    >>> import bifabrik as bif
    >>>
    >>> df = spark.read.format("csv").option("header","true").load("Files/CsvFiles/annual-enterprise-survey-2021.csv")
    >>> bif.fromSparkDf(df).toTable('Table1').run()
    """
    ds = SparkDfSource(__prepPipeline(), df)
    return ds

def fromPandasDf(df: PandasDf) -> PandasDfSource:
    """Use pandas dataframe as source
    
    Examples
    --------
    
    >>> import bifabrik as bif
    >>>
    >>> df = pd.read_csv(...)
    >>> bif.fromPandasDf(df).toTable('Table1').run()
    """
    ds = PandasDfSource(__prepPipeline(), df)
    return ds

def fromSharePointList(siteUrl: str, listName: str) -> SharePointListSource:
    """Loads data from an Office365 SharePoint list. You will need to configure the credentials for SharePoint.
    
    Examples
    --------
    
    >>> import bifabrik as bif
    >>> 
    >>> bif.config.security.keyVaultUrl = 'https://kv-fabrik.vault.azure.net/'
    >>> bif.config.security.loginKVSecretName = 'SharePointLogin'
    >>> bif.config.security.passwordKVSecretName = 'SharePointPwd'
    >>> 
    >>> bif.fromSharePointList('https://fabrik.sharepoint.com/sites/BusinessIntelligence', 'CustomerList') \
    >>>     .toTable('Customers').run()
    """
    ds = SharePointListSource(__prepPipeline(), siteUrl, listName)
    return ds

def __getattr__(name):
    if name == 'config':
        return __configuration
    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")



from bifabrik.tsf.DataTransformation import DataTransformation
from bifabrik.cfg.ValidationTransformationConfiguration import ValidationTransformationConfiguration
import bifabrik.utils.log as lg
from pyspark.sql.functions import col
from pyspark.sql.dataframe import DataFrame as SparkDf

class ValidationTransformation(DataTransformation, ValidationTransformationConfiguration):
    """Test the input data by checking the ValidationResult and ValidationMessage columns to determine if each row failed / passed the validation.
    The column names / values checked can be changed in configuration.
    
    Examples
    --------
    
    >>> import bifabrik as bif
    >>>
    >>> bif.fromSql('''
    >>> WITH values AS(
    >>> SELECT AnnualSurveyID
    >>> ,Variable_code, Value
    >>> ,CAST(REPLACE(Value, ',', '') AS DOUBLE) ValueDouble 
    >>> FROM AnnualSurveyFromDW1
    >>> )
    >>> SELECT v.*
    >>> ,IF(v.ValueDouble IS NULL, 'Error', 'OK') AS ValidationResult
    >>> ,IF(v.ValueDouble IS NULL, CONCAT('"', Value, '" cannot be converted to double'), 'OK') AS ValidationMessage
    >>> FROM values v
    >>> ''').validate('NumberParsingTest') \
    >>> .toTable('SurveyValuesParsed').run()
    """

    def __init__(self, parentPipeline, testName = 'Unnamed'):
        super().__init__(parentPipeline)
        ValidationTransformationConfiguration.__init__(self)
        self.validation.testName = testName
    
    def __str__(self):
        return f'Data validation \'{self.validation.testName}\''
    
    def execute(self, input) -> SparkDf:
        lgr = lg.getLogger()
        self.__logger = lgr
        self._error = None
        
        mergedConfig = self._pipeline.configuration.mergeToCopy(self)
        config = mergedConfig.validation

        resultColumnName = config.resultColumnName
        messageColumnName = config.messageColumnName
        errorResultValue = config.errorResultValue
        warningResultValue = config.warningResultValue
        # okResultValue = config.okResultValue
        # fail / log / None
        onError = config.onError.lower()
        onWarning = config.onWarning.lower()
        self.__resultColumnName = resultColumnName
        self.__messageColumnName = messageColumnName

        maxErrors = config.maxErrors
        maxWarnings = config.maxWarnings

        if maxErrors < 1:
            raise Exception('Invalid configuration - maxErrors has to be a positive integer')
        if maxWarnings < 1:
            raise Exception('Invalid configuration - maxWarnings has to be a positive integer')

        if self.validation.testName == None:
            self.validation.testName = ''
        self.__testName = self.validation.testName

        # resultColumnName = 'ValidationResult'
        # messageColumnName = 'ValidationMessage'
        # errorResultValue = 'Error'
        # warningResultValue = 'Warning'
        # okResultValue = 'OK'
        # onError = 'fail'
        # onWarning = 'log'
        # testName = None
        
        if not (resultColumnName in input.columns):
            raise Exception(f'The input dataframe does not contain the result column `{resultColumnName}`')
        if not (messageColumnName in input.columns):
            raise Exception(f'The input dataframe does not contain the result column `{messageColumnName}`')
        
        errors = input.filter(col(resultColumnName) == errorResultValue).collect()
        warnings = input.filter(col(resultColumnName) == warningResultValue).collect()
        oks = input.filter(col(resultColumnName) != errorResultValue).filter(col(resultColumnName) != warningResultValue)

        if onError is None:
            onError = 'none'
        if onWarning is None:
            onWarning = 'none'
        
        # log
        if onError in ['fail', 'log']:
            remaining = maxErrors
            for error in errors:
                if remaining < 1:
                    break
                err = f'Test {self.__testName}: {error[self.__messageColumnName]}; {self.rowToStr(error)}'
                lgr.error(err)
                print(f'Error: {err}')
                remaining = remaining - 1
        
        if onWarning in ['fail', 'log']:
            remaining = maxWarnings
            for warning in warnings:
                if remaining < 1:
                    break
                wrn = f'Test {self.__testName}: {warning[self.__messageColumnName]}; {self.rowToStr(warning)}'
                lgr.warning(wrn)
                print(f'Warning: {wrn}')
                remaining = remaining - 1

        # fail if need be
        if onError == 'fail' and len(errors) > 0:
            error = errors[0]
            raise Exception(f'Test {self.__testName} failed: {error[self.__messageColumnName]}; {self.rowToStr(error)}')
        
        if onWarning == 'fail' and len(warnings) > 0:
            warning = warnings[0]
            raise Exception(f'Test {self.__testName} failed: {warning[self.__messageColumnName]}; {self.rowToStr(warning)}')
        
        self._result = input
        self._completed = True
        return self._result
    
    def rowToStr(self, row) -> str:
        d = row.asDict()
        r = ''
        for k in d:
            if k == self.__resultColumnName:
                continue
            if k == self.__messageColumnName:
                continue
            r += f'`{k}`: {d[k]}\t'
        return r


from bifabrik.tsf.DataTransformation import DataTransformation
from pyspark.sql.dataframe import DataFrame as SparkDf
from pandas.core.frame import DataFrame as PandasDf

class PandasDfTransformation(DataTransformation):
    """Use spark dataframe as source
    
    Examples
    --------
    
    >>> def tf(df):
    >>>     df.insert(0, 'NewColumn', 'NewValue')
    >>>     return df
    >>>
    >>> (
    >>> bif
    >>>     .fromCsv("Files/CsvFiles/annual-enterprise-survey-2021.csv")
    >>>     .transformPandasDf(tf)
    >>>     .toTable("Survey3")
    >>>     .run()
    >>> )
    """

    def __init__(self, parentPipeline, func):
        super().__init__(parentPipeline)
        self.__func = func
    
    def __str__(self):
        return f'Pandas DF transformation: {self.__func}'
    
    def execute(self, input) -> SparkDf:
        pandasInput =  input.toPandas()
        pandasResult = self.__func(pandasInput)
        self._result = self._spark.createDataFrame(pandasResult)
        self._completed = True
        return self._result


from bifabrik.cfg.engine.ConfigContainer import ConfigContainer
from bifabrik.cfg.specific.ExcelConfiguration import ExcelConfiguration
from bifabrik.cfg.specific.FileSourceConfiguration import FileSourceConfiguration
from bifabrik.cfg.DataSourceConfigurationBase import DataSourceConfigurationBase

class ExcelSourceConfiguration(DataSourceConfigurationBase):

    def __init__(self):
        self.__excel = ExcelConfiguration()
        self.__fileSource = FileSourceConfiguration()
        super().__init__()
        

    @property
    def excel(self) -> ExcelConfiguration:
        return self.__excel
    
    @property
    def fileSource(self) -> FileSourceConfiguration:
        return self.__fileSource


"""
Delta table utilities - add / rename / remove columns
"""

import pyspark.sql.session as pss
from pyspark.sql.functions import lit, col
import bifabrik.utils.fsUtils as fsu
from datetime import datetime

spark = pss.SparkSession.builder.getOrCreate()

def addTableColumnFromValue(databaseName: str, tableName: str, columnName: str, value):
    """Adds a new column to the delta table with the provided value.
    The `database` parameter refers to a lakehouse in the current workspace
    """
    # find out the data type
    df = spark.createDataFrame([('0','0')], ['____', '_____'])
    df = df.withColumn(columnName, lit(value))
    dts = df.dtypes
    dfColT = dts[2]
    type_name = dfColT[1]
    alterQuery = f'ALTER TABLE `{databaseName}`.`{tableName}` ADD columns (`{columnName}` {type_name})'
    updateQuery = f'UPDATE `{databaseName}`.`{tableName}` SET `{columnName}` = CAST(\'{value}\' AS {type_name})'

    print(alterQuery)
    spark.sql(alterQuery)
    print(updateQuery)
    spark.sql(updateQuery)

def addTableColumnFromType(databaseName: str, tableName: str, columnName: str, typeName: str):
    """Adds a new empty column to the delta table.
    The `database` parameter refers to a lakehouse in the current workspace
    """
    alterQuery = f'ALTER TABLE `{databaseName}`.`{tableName}` ADD columns (`{columnName}` {typeName})'
    print(alterQuery)
    spark.sql(alterQuery)

def renameTableColumn(databaseName: str, tableName: str, columnName: str, newColumnName: str):
    """Renames a column in a delta table"""
    alterPropsQuery = f'''ALTER TABLE `{databaseName}`.`{tableName}` SET TBLPROPERTIES (
    'delta.minReaderVersion' = '2',
    'delta.minWriterVersion' = '5',
    'delta.columnMapping.mode' = 'name'
    )
    '''
    print(alterPropsQuery)
    spark.sql(alterPropsQuery)

    alterQuery = f'ALTER TABLE `{databaseName}`.`{tableName}` RENAME COLUMN `{columnName}` TO `{newColumnName}`'
    print(alterQuery)
    spark.sql(alterQuery)

def dropTableColumn(databaseName: str, tableName: str, columnName: str):
    """Drops a column from a delta table"""
    alterPropsQuery = f'''ALTER TABLE `{databaseName}`.`{tableName}` SET TBLPROPERTIES (
    'delta.minReaderVersion' = '2',
    'delta.minWriterVersion' = '5',
    'delta.columnMapping.mode' = 'name'
    )
    '''
    print(alterPropsQuery)
    spark.sql(alterPropsQuery)

    alterQuery = f'ALTER TABLE `{databaseName}`.`{tableName}` DROP COLUMN `{columnName}`'
    print(alterQuery)
    spark.sql(alterQuery)


def listTables():
    """List table names in the current lakehouse"""
    currentLh = fsu.currentLakehouse()
    if currentLh is None:
        return []
    
    df = spark.sql('SHOW TABLES')
    df = df.filter('isTemporary = False').select('tableName')
    l = list(map(lambda x: x.tableName, df.collect()))

    dfw = spark.sql('SHOW VIEWS')
    lw = list(map(lambda x: x.viewName, dfw.collect()))

    res = []
    for element in l:
        if element not in lw:
            res.append(element)
    
    return res

'''The delta table history for all tables in the list, returned as a single dataframe

Examples
--------

>>> import bifabrik.utils.tableUtils as tu
>>> 
>>> tableList = tu.listTables()
>>> th = tu.tablesHistory(tableList)
'''
def tablesHistory(tables):
    '''Lists the history for all tables in the list in one dataframe'''
    uni_df = None
    for tbl in tables:
        df = spark.sql(f'DESCRIBE HISTORY {tbl}')
        df = df.select(lit(tbl).alias("tableName"), "*")
        if uni_df is None:
            uni_df = df
        uni_df = uni_df.union(df)
    return uni_df

def restoreToPIT(tables, timestamp):
    '''Restores the tables in the current lakehouse to the point in time (their last version before the given timestamp)

    Parameters:
        tables (array[str]):    list of table names (['table1', 'table2']). You can get the list of all tables using listTables()
        timestamp:  Either a string in the format '%Y-%m-%d %H:%M:%S.%f' (UTC time) or datetime.datetime

    Examples
    --------

    >>> import bifabrik.utils.tableUtils as tu
    >>> tables = tu.listTables()
    >>> tu.restoreToPIT(tables, '2024-05-08 16:30:00')

    '''
    ts = None
    if isinstance(timestamp, datetime):
        ts = timestamp
    elif isinstance(timestamp, str):
        if '.' in timestamp:
            ts = datetime.strptime(timestamp, '%Y-%m-%d %H:%M:%S.%f')
        else:
            ts = datetime.strptime(timestamp, '%Y-%m-%d %H:%M:%S')
    
    history = tablesHistory(tables)
    eq = history.where(f"timestamp = '{ts}'")
    if eq.count() > 0:
        raise Exception(f"There are transactions with the exact timestamp '{ts}'. Please select a point in time between the transactions.")
    
    prev = history.where(f"timestamp < '{ts}'")
    for t in tables:
        tprev = prev.filter(f"tableName = '{t}'")
        if tprev.count() == 0:
            print(f"`{t}` has no versions before {ts}; skipping")
            continue
        h = tprev.orderBy(tprev.timestamp.desc()).head()
        print(f"Restoring `{h.tableName}` to version {h.version} ({h.timestamp})")
        rsql = f"RESTORE TABLE `{h.tableName}` TO VERSION AS OF {h.version}"
        spark.sql(rsql)
        



from bifabrik.src.DataSource import DataSource
from pandas.core.frame import DataFrame as PandasDf
from pyspark.sql.dataframe import DataFrame as SparkDf

class PandasDfSource(DataSource):
    """Use pandas dataframe as source
    
    Examples
    --------
    
    >>> import bifabrik as bif
    >>>
    >>> df = pd.read_csv(...)
    >>> bif.fromPandasDf(df).toTable('Table1').run()
    """

    def __init__(self, parentPipeline, df: PandasDf):
        if not isinstance(df, PandasDf):
            raise Exception('The source is not a pandas dataframe')
        
        super().__init__(parentPipeline)
        self.__pandasDf = df
    
    def __str__(self):
        return f'Pandas DF source: {self.__pandasDf}'
    
    def execute(self, input) -> SparkDf:
        self._result = self._spark.createDataFrame(self.__pandasDf)
        self._completed = True
        return self._result


from bifabrik.cfg.engine.Configuration import Configuration
from bifabrik.cfg.engine.Configuration import CfgProperty

class MetadataStorageConfiguration(Configuration):
    """Defines the data metadata lakehouse. By default refers to the default lakehouse of the notebook. This is used for logs and metadata tables
    """
    def __init__(self):
        self._explicitProps = {}
        self.__metadataWorkspaceName = None
        self.__metadataLakehouseName = None
        self.__tmslBackupPathPattern = 'Files/tmsl_backup/{workspacename}/{datasetname}/{datasetname}_{timestamp}.json'

    @CfgProperty
    def metadataWorkspaceName(self) -> str:
        """The workspace where bifabrik stores metadata
        """
        return self.__metadataWorkspaceName
    @metadataWorkspaceName.setter(key='metadataWorkspaceName')
    def metadataWorkspaceName(self, val):
        self.__metadataWorkspaceName = val

    @CfgProperty
    def metadataLakehouseName(self) -> str:
        """The lakehouse where bifabrik stores metadata
        """
        return self.__metadataLakehouseName
    @metadataLakehouseName.setter(key='metadataLakehouseName')
    def metadataLakehouseName(self, val):
        self.__metadataLakehouseName = val
    
    @CfgProperty
    def tmslBackupPathPattern(self) -> str:
        """The path pattern where to save backups of semantic model definitions (in TMSL format). 
        
        Supported placeholders:
            {workspacename} : the source workspace name
            {datasetname}   : the source dataset name
            {timestamp}     : backup timestamp as %Y_%m_%d_%H_%M_%S_%f
        
        default: 'Files/tmsl_backup/{workspacename}/{datasetname}/{datasetname}_{timestamp}.json'

        The metadataLakehouseName / metadataWorkspaceName should also be configured. If not, the file will be saved to the default lakehouse attached to the notebook.

        """
        return self.__tmslBackupPathPattern
    @tmslBackupPathPattern.setter(key='tmslBackupPathPattern')
    def tmslBackupPathPattern(self, val):
        self.__tmslBackupPathPattern = val
    
